"""Handle global package state."""
from __future__ import absolute_import

from thriftworker.utils.proxy import Proxy

default_cls = None
default_app = None


def as_default_cls(cls):
    global default_cls
    default_cls = cls
    return cls


def get_default_cls():
    global default_cls
    if default_cls is None:
        from thriftpool.app.base import ThriftPool
        return ThriftPool
    return default_cls


def set_current_app(app):
    global default_app
    default_app = app


def get_current_app():
    global default_app
    if default_app is None:
        # creates the default app, but we want to defer that.
        cls = get_default_cls()
        set_current_app(cls())
    return default_app


current_app = Proxy(get_current_app)



from __future__ import absolute_import

import six

__all__ = ['smart_str', 'smart_unicode']


def smart_str(s, encoding='utf-8', strings_only=False, errors='strict'):
    """Returns a bytestring version of 's', encoded as specified in 'encoding'.

    If strings_only is True, don't convert (some) non-string-like objects.

    """
    if strings_only and (s is None or isinstance(s, six.integer_types)):
        return s
    elif not isinstance(s, six.string_types):
        try:
            return str(s)
        except UnicodeEncodeError:
            if isinstance(s, Exception):
                # An Exception subclass containing non-ASCII data that doesn't
                # know how to print itself properly. We shouldn't raise a
                # further exception.
                return ' '.join([smart_str(arg, encoding, strings_only,
                                           errors) for arg in s])
            return unicode(s).encode(encoding, errors)
    elif isinstance(s, unicode):
        return s.encode(encoding, errors)
    elif s and encoding != 'utf-8':
        return s.decode('utf-8', errors).encode(encoding, errors)
    else:
        return s


def smart_unicode(s, encoding='utf-8', strings_only=False, errors='strict'):
    """Returns a unicode object representing 's'. Treats bytestrings using the
    'encoding' codec.

    If strings_only is True, don't convert (some) non-string-like objects.

    """
    # Handle the common case first, saves 30-40% in performance when s
    # is an instance of unicode. This function gets called often in that
    # setting.
    if isinstance(s, unicode):
        return s
    if not isinstance(s, six.string_types):
        if hasattr(s, '__unicode__'):
            s = unicode(s)
        else:
            try:
                s = unicode(str(s), encoding, errors)
            except UnicodeEncodeError:
                if not isinstance(s, Exception):
                    raise
                # If we get to here, the caller has passed in an Exception
                # subclass populated with non-ASCII data without special
                # handling to display as a string. We need to handle this
                # without raising a further exception. We do an
                # approximation to what the Exception's standard str()
                # output should be.
                s = ' '.join([smart_unicode(arg, encoding, strings_only,
                                            errors) for arg in s])
    elif not isinstance(s, unicode):
        # Note: We use .decode() here, instead of unicode(s, encoding,
        # errors), so that if s is a SafeString, it ends up being a
        # SafeUnicode at the end.
        s = s.decode(encoding, errors)
    return s



from __future__ import absolute_import

import os
import cPickle as pickle
from struct import Struct, calcsize
from select import select


class StreamSerializer(object):
    """Helper to pass python objects over streams."""

    length_format = '!i'

    def __init__(self):
        self.length_struct = Struct(self.length_format)
        self.length = calcsize(self.length_format)

    @staticmethod
    def encode(obj):
        return pickle.dumps(obj)

    @staticmethod
    def decode(message):
        return pickle.loads(message)

    def encode_with_length(self, obj):
        """Encode object and prepend length to message."""
        message = self.encode(obj)
        return self.length_struct.pack(len(message)) + message

    def decode_from_stream(self, fd, timeout=5):
        """Read object from given stream and return it."""
        rlist, _, _ = select([fd], [], [], timeout)
        if not rlist:
            raise RuntimeError("Can't read object from {0!r}.".format(fd))
        message_length = self.length_struct.unpack(os.read(fd, self.length))[0]
        assert message_length > 0, 'wrong message length provided'
        return self.decode(os.read(fd, message_length))



"""Some base structures.

This file was copied and adapted from celery.

:copyright: (c) 2009 - 2012 by Ask Solem.
:license: BSD, see LICENSE for more details.

"""
from __future__ import absolute_import

from collections import defaultdict
import itertools

from six import iteritems, iterkeys

__all__ = ['DependencyGraph', 'AttributeDict', 'AggregatedView']


class DependencyGraph(object):
    """A directed acyclic graph of objects and their dependencies.

    Supports a robust topological sort
    to detect the order in which they must be handled.

    Takes an optional iterator of ``(obj, dependencies)``
    tuples to build the graph from.

    .. warning::

        Does not support cycle detection.

    """

    def __init__(self, it=None):
        self.adjacent = {}
        if it is not None:
            self.update(it)

    def add_arc(self, obj):
        """Add an object to the graph."""
        self.adjacent.setdefault(obj, [])

    def add_edge(self, A, B):
        """Add an edge from object ``A`` to object ``B``
        (``A`` depends on ``B``)."""
        self[A].append(B)

    def topsort(self):
        """Sort the graph topologically.

        :returns: a list of objects in the order
            in which they must be handled.

        """
        graph = DependencyGraph()
        components = self._tarjan72()

        NC = dict((node, component)
                  for component in components
                  for node in component)
        for component in components:
            graph.add_arc(component)
        for node in self:
            node_c = NC[node]
            for successor in self[node]:
                successor_c = NC[successor]
                if node_c != successor_c:
                    graph.add_edge(node_c, successor_c)
        return [t[0] for t in graph._khan62()]

    def valency_of(self, obj):
        """Returns the velency (degree) of a vertex in the graph."""
        try:
            l = [len(self[obj])]
        except KeyError:
            return 0
        for node in self[obj]:
            l.append(self.valency_of(node))
        return sum(l)

    def update(self, it):
        """Update the graph with data from a list
        of ``(obj, dependencies)`` tuples."""
        tups = list(it)
        for obj, _ in tups:
            self.add_arc(obj)
        for obj, deps in tups:
            for dep in deps:
                self.add_edge(obj, dep)

    def edges(self):
        """Returns generator that yields for all edges in the graph."""
        return (obj for obj, adj in self.iteritems() if adj)

    def _khan62(self):
        """Khans simple topological sort algorithm from '62

        See http://en.wikipedia.org/wiki/Topological_sorting

        """
        count = defaultdict(lambda: 0)
        result = []

        for node in self:
            for successor in self[node]:
                count[successor] += 1
        ready = [node for node in self if not count[node]]

        while ready:
            node = ready.pop()
            result.append(node)

            for successor in self[node]:
                count[successor] -= 1
                if count[successor] == 0:
                    ready.append(successor)
        result.reverse()
        return result

    def _tarjan72(self):
        """Tarjan's algorithm to find strongly connected components.

        See http://bit.ly/vIMv3h.

        """
        result, stack, low = [], [], {}

        def visit(node):
            if node in low:
                return
            num = len(low)
            low[node] = num
            stack_pos = len(stack)
            stack.append(node)

            for successor in self[node]:
                visit(successor)
                low[node] = min(low[node], low[successor])

            if num == low[node]:
                component = tuple(stack[stack_pos:])
                stack[stack_pos:] = []
                result.append(component)
                for item in component:
                    low[item] = len(self)

        for node in self:
            visit(node)

        return result

    def to_dot(self, fh, ws=" " * 4):
        """Convert the graph to DOT format.

        :param fh: A file, or a file-like object to write the graph to.

        """
        fh.write("digraph dependencies {\n")
        for obj, adjacent in self.iteritems():
            if not adjacent:
                fh.write(ws + '"%s"\n' % (obj,))
            for req in adjacent:
                fh.write(ws + '"%s" -> "%s"\n' % (obj, req))
        fh.write("}\n")

    def __iter__(self):
        return iterkeys(self.adjacent)

    def __getitem__(self, node):
        return self.adjacent[node]

    def __len__(self):
        return len(self.adjacent)

    def __contains__(self, obj):
        return obj in self.adjacent

    def _iterate_items(self):
        return iteritems(self.adjacent)
    items = iteritems = _iterate_items

    def __repr__(self):
        return '\n'.join(self.repr_node(N) for N in self)

    def repr_node(self, obj, level=1):
        output = ["%s(%s)" % (obj, self.valency_of(obj))]
        if obj in self:
            for other in self[obj]:
                d = "%s(%s)" % (other, self.valency_of(other))
                output.append('     ' * level + d)
                output.extend(self.repr_node(other, level + 1).split('\n')[1:])
        return '\n'.join(output)


class AttributeDictMixin(object):
    """Adds attribute access to mappings.

    `d.key -> d[key]`

    """

    def __getattr__(self, k):
        """`d.key -> d[key]`"""
        try:
            return self[k]
        except KeyError:
            raise AttributeError(
                "'%s' object has no attribute '%s'" % (type(self).__name__, k))

    def __setattr__(self, key, value):
        """`d[key] = value -> d.key = value`"""
        self[key] = value


class AttributeDict(dict, AttributeDictMixin):
    """Dict subclass with attribute access."""
    pass


class AggregatedView(AttributeDictMixin):
    """Combine multiple dictionary to one."""

    def __init__(self, d):
        changes = {}
        self.__dict__.update(_changes=changes,
                             _underlying_dicts=[changes, d])

    def add_default(self, d):
        self._underlying_dicts.insert(2, d)

    def __getitem__(self, key):
        for d in self._underlying_dicts:
            try:
                return d[key]
            except KeyError:
                pass
        raise KeyError(key)

    def __setitem__(self, key, value):
        self._changes[key] = value

    def __contains__(self, key):
        for d in self._underlying_dicts:
            if key in d:
                return True
        return False

    def __iter__(self):
        return iter(set(itertools.chain.from_iterable(
                        [self._changes] + self._underlying_dicts)))

    def get(self, key, default=None):
        try:
            return self[key]
        except KeyError:
            return default

    def setdefault(self, key, default):
        try:
            return self[key]
        except KeyError:
            self[key] = default
            return default

    def update(self, *args, **kwargs):
        return self._changes.update(*args, **kwargs)

    def items(self):
        return [(key, self[key]) for key in self]

    def values(self):
        return [self[key] for key in self]

    def keys(self):
        return list(self)

    def __repr__(self):
        return repr(dict(self.items()))



from __future__ import absolute_import

from logging import getLogger

from six import iteritems

from thriftpool.utils.platforms import set_process_title
from thriftpool.components.base import Namespace
from thriftpool.controllers.base import Controller

logger = getLogger(__name__)


class WorkerNamespace(Namespace):

    name = 'worker'

    def modules(self):
        return ['thriftpool.components.loop',
                'thriftpool.components.worker',
                'thriftpool.components.broker']


class WorkerController(Controller):

    Namespace = WorkerNamespace

    ignore_interrupt = True
    acceptors = None

    def __init__(self, start_fd):
        self.handshake_fd = start_fd
        self.outgoing_fd = self.handshake_fd + 1
        self.incoming_fd = self.outgoing_fd + 1
        super(WorkerController, self).__init__()

    def on_before_init(self):
        super(WorkerController, self).on_before_init()
        app = self.app
        if app.config.REDIRECT_STDOUT:
            logger = getLogger('thriftpool.stdout')
            app.log.redirect_stdouts_to_logger(logger)

    def change_title(self, name):
        """Change process title."""
        self._debug('Change process title to %r.', name)
        set_process_title(name)

    def register_acceptors(self, descriptors):
        """Register all existed acceptors with given descriptors."""
        acceptors = self.acceptors
        slots = self.app.slots
        delta = self.incoming_fd + 1
        for fd, name in iteritems(descriptors):
            slot = slots[name]
            fd += delta
            self._debug('Register acceptor %r with fd %d.', name, fd)
            acceptors.register(fd, name, backlog=slot.listener.backlog)

    def start_acceptor(self, name):
        """Start acceptors by it's name."""
        self._debug('Start acceptor %r.', name)
        self.acceptors.start_by_name(name)

    def stop_acceptor(self, name):
        """Stop acceptors by it's name."""
        self._debug('Stop acceptor %r.', name)
        self.acceptors.stop_by_name(name)

    def get_stack(self):
        """Return currently running methods."""
        return self.app.request_stack.to_dict()

    def get_counters(self):
        """Return counters here."""
        return self.app.thriftworker.counters.to_dict()

    def get_timers(self):
        """Return timers here."""
        return self.app.thriftworker.timers.to_dict()



#
# Autogenerated by Thrift Compiler (0.9.0)
#
# DO NOT EDIT UNLESS YOU ARE SURE THAT YOU KNOW WHAT YOU ARE DOING
#
#  options string: py:new_style,utf8strings,slots,dynamic
#

from thrift.Thrift import TType, TMessageType, TException, TApplicationException
from ttypes import *
from thrift.Thrift import TProcessor
from thrift.protocol.TBase import TBase, TExceptionBase


class Iface(object):
  def ping(self, ):
    pass

  def echoString(self, s):
    """
    Parameters:
     - s
    """
    pass


class Client(Iface):
  def __init__(self, iprot, oprot=None):
    self._iprot = self._oprot = iprot
    if oprot is not None:
      self._oprot = oprot
    self._seqid = 0

  def ping(self, ):
    self.send_ping()
    self.recv_ping()

  def send_ping(self, ):
    self._oprot.writeMessageBegin('ping', TMessageType.CALL, self._seqid)
    args = ping_args()
    args.write(self._oprot)
    self._oprot.writeMessageEnd()
    self._oprot.trans.flush()

  def recv_ping(self, ):
    (fname, mtype, rseqid) = self._iprot.readMessageBegin()
    if mtype == TMessageType.EXCEPTION:
      x = TApplicationException()
      x.read(self._iprot)
      self._iprot.readMessageEnd()
      raise x
    result = ping_result()
    result.read(self._iprot)
    self._iprot.readMessageEnd()
    return

  def echoString(self, s):
    """
    Parameters:
     - s
    """
    self.send_echoString(s)
    return self.recv_echoString()

  def send_echoString(self, s):
    self._oprot.writeMessageBegin('echoString', TMessageType.CALL, self._seqid)
    args = echoString_args()
    args.s = s
    args.write(self._oprot)
    self._oprot.writeMessageEnd()
    self._oprot.trans.flush()

  def recv_echoString(self, ):
    (fname, mtype, rseqid) = self._iprot.readMessageBegin()
    if mtype == TMessageType.EXCEPTION:
      x = TApplicationException()
      x.read(self._iprot)
      self._iprot.readMessageEnd()
      raise x
    result = echoString_result()
    result.read(self._iprot)
    self._iprot.readMessageEnd()
    if result.success is not None:
      return result.success
    raise TApplicationException(TApplicationException.MISSING_RESULT, "echoString failed: unknown result");


class Processor(Iface, TProcessor):
  def __init__(self, handler):
    self._handler = handler
    self._processMap = {}
    self._processMap["ping"] = Processor.process_ping
    self._processMap["echoString"] = Processor.process_echoString

  def process(self, iprot, oprot):
    (name, type, seqid) = iprot.readMessageBegin()
    if name not in self._processMap:
      iprot.skip(TType.STRUCT)
      iprot.readMessageEnd()
      x = TApplicationException(TApplicationException.UNKNOWN_METHOD, 'Unknown function %s' % (name))
      oprot.writeMessageBegin(name, TMessageType.EXCEPTION, seqid)
      x.write(oprot)
      oprot.writeMessageEnd()
      oprot.trans.flush()
      return
    else:
      self._processMap[name](self, seqid, iprot, oprot)
    return True

  def process_ping(self, seqid, iprot, oprot):
    args = ping_args()
    args.read(iprot)
    iprot.readMessageEnd()
    result = ping_result()
    self._handler.ping()
    oprot.writeMessageBegin("ping", TMessageType.REPLY, seqid)
    result.write(oprot)
    oprot.writeMessageEnd()
    oprot.trans.flush()

  def process_echoString(self, seqid, iprot, oprot):
    args = echoString_args()
    args.read(iprot)
    iprot.readMessageEnd()
    result = echoString_result()
    result.success = self._handler.echoString(args.s)
    oprot.writeMessageBegin("echoString", TMessageType.REPLY, seqid)
    result.write(oprot)
    oprot.writeMessageEnd()
    oprot.trans.flush()


# HELPER FUNCTIONS AND STRUCTURES

class ping_args(TBase):

  __slots__ = [ 
   ]

  thrift_spec = (
  )


class ping_result(TBase):

  __slots__ = [ 
   ]

  thrift_spec = (
  )


class echoString_args(TBase):
  """
  Attributes:
   - s
  """

  __slots__ = [ 
    's',
   ]

  thrift_spec = (
    None, # 0
    (1, TType.STRING, 's', None, None, ), # 1
  )

  def __init__(self, s=None,):
    self.s = s


class echoString_result(TBase):
  """
  Attributes:
   - success
  """

  __slots__ = [ 
    'success',
   ]

  thrift_spec = (
    (0, TType.STRING, 'success', None, None, ), # 0
  )

  def __init__(self, success=None,):
    self.success = success




# -*- coding: utf-8 -*-
"""Implements components and boot-steps.

This file was copied and adapted from celery.

:copyright: (c) 2009 - 2012 by Ask Solem.
:license: BSD, see LICENSE for more details.

"""
from __future__ import absolute_import

from collections import defaultdict
from importlib import import_module
from logging import getLogger

from six import with_metaclass

from thriftworker.utils.imports import instantiate

from thriftpool.utils.mixin import LogsMixin
from thriftpool.utils.structures import DependencyGraph

logger = getLogger(__name__)


class Namespace(LogsMixin):
    """A namespace containing components.

    Every component must belong to a namespace.

    When component classes are created they are added to the
    mapping of unclaimed components.  The components will be
    claimed when the namespace they belong to is created.

    :keyword name: Set the name of this namespace.
    :keyword app: Set the Celery app for this namespace.

    """
    name = None
    _unclaimed = defaultdict(dict)
    _started_count = 0

    def __init__(self, name=None, app=None):
        self.app = app
        self.name = name or self.name
        self.services = []

    def modules(self):
        """Subclasses can override this to return a
        list of modules to import before components are claimed."""
        return []

    def load_modules(self):
        """Will load the component modules this namespace depends on."""
        for m in self.modules():
            self.import_module(m)

    def apply(self, parent, **kwargs):
        """Apply the components in this namespace to an object.

        This will apply the ``__init__`` and ``include`` methods
        of each components with the object as argument.

        For ``StartStopComponents`` the services created
        will also be added the the objects ``components`` attribute.

        """
        self._debug("Loading modules.")
        self.load_modules()
        self._debug("Claiming components.")
        self.components = self._claim()
        self._debug("Building boot step graph.")
        self.boot_steps = [self.bind_component(name, parent, **kwargs)
                           for name in self._finalize_boot_steps()]
        self._debug("New boot order: {%s}",
                    ', '.join(c.name for c in self.boot_steps))

        for component in self.boot_steps:
            component.include(parent)
        return self

    def bind_component(self, name, parent, **kwargs):
        """Bind component to parent object and this namespace."""
        comp = self[name](parent, **kwargs)
        comp.namespace = self
        return comp

    def import_module(self, module):
        return import_module(module)

    def __getitem__(self, name):
        return self.components[name]

    def _find_last(self):
        for C in self.components.itervalues():
            if C.last:
                return C

    def _finalize_boot_steps(self):
        G = self.graph = \
            DependencyGraph((C.name, C.requires)
                            for C in self.components.itervalues())
        last = self._find_last()
        if last:
            for obj in G:
                if obj != last.name:
                    G.add_edge(last.name, obj)
        return G.topsort()

    def _claim(self):
        return self._unclaimed[self.name]


class ComponentType(type):
    """Metaclass for components."""

    def __new__(cls, name, bases, attrs):
        if name == 'NewBase':
            return super(ComponentType, cls).__new__(cls, name, bases, attrs)
        abstract = attrs.pop("abstract", False)
        if not abstract:
            try:
                cname = attrs["name"]
            except KeyError:
                raise NotImplementedError("Components must be named")
            namespace = attrs.get("namespace", None)
            if not namespace:
                attrs["namespace"], _, attrs["name"] = cname.partition('.')
        cls = super(ComponentType, cls).__new__(cls, name, bases, attrs)
        if not abstract:
            Namespace._unclaimed[cls.namespace][cls.name] = cls
        return cls


class Component(with_metaclass(ComponentType)):
    """A component.

    The :meth:`__init__` method is called when the component
    is bound to a parent object, and can as such be used
    to initialize attributes in the parent object at
    parent instantiation-time.

    """

    #: The name of the component, or the namespace
    #: and the name of the component separated by dot.
    name = None

    #: List of component names this component depends on.
    #: Note that the dependencies must be in the same namespace.
    requires = ()

    #: can be used to specify the namespace,
    #: if the name does not include it.
    namespace = None

    #: if set the component will not be registered,
    #: but can be used as a component base class.
    abstract = True

    #: Optional obj created by the :meth:`create` method.
    #: This is used by StartStopComponents to keep the
    #: original service object.
    obj = None

    #: This flag is reserved for the workers Consumer,
    #: since it is required to always be started last.
    #: There can only be one object marked with last
    #: in every namespace.
    last = False

    #: This provides the default for :meth:`include_if`.
    enabled = True

    def __init__(self, parent, **kwargs):
        pass

    def create(self, parent):
        """Create the component."""
        raise NotImplementedError('subclass responsibility')

    def include_if(self, parent):
        """An optional predicate that decided whether this
        component should be created."""
        return self.enabled

    def instantiate(self, qualname, *args, **kwargs):
        return instantiate(qualname, *args, **kwargs)

    def include(self, parent):
        if self.include_if(parent):
            self.obj = self.create(parent)
            return True


class StartStopComponent(Component):
    abstract = True
    terminable = False

    def start(self):
        if self.obj is None:
            return
        self.obj.start()

    def stop(self):
        if self.obj is None:
            return
        self.obj.stop()

    def terminate(self):
        if self.obj is None:
            return
        if self.terminable:
            self.obj.terminate()
        self.obj.stop()

    def include(self, parent):
        if super(StartStopComponent, self).include(parent):
            parent.components.append(self)



"""Implement base class for handlers. This can be used to simply inherit from
this class instead of directly registering new handler.

"""
from __future__ import absolute_import

from six import with_metaclass

from thriftpool import thriftpool
from thriftpool.exceptions import RegistrationError
from thriftpool.utils.structures import AttributeDict

__all__ = ['BaseHandler']


class Options(AttributeDict):
    """Holder for handler options."""


class OptionsMeta(type):
    """Create options for class."""

    def __new__(mcs, name, bases, attrs):
        # Create default options for new class.
        attrs['meta'] = meta = Options(abstract=False, enabled=True)
        # Try to inherit from parent class.
        for base in bases[::-1]:
            try:
                meta.update(base.meta)
            except AttributeError:
                pass
        meta.abstract = False
        meta.enabled = True
        # Fill options.
        try:
            options = attrs['options']
        except KeyError:
            pass
        else:
            meta.update({key: getattr(options, key)
                         for key in dir(options)
                         if not key.startswith('_')})
        # Create new class itself.
        return super(OptionsMeta, mcs).__new__(mcs, name, bases, attrs)


class HandlerMeta(OptionsMeta):
    """Register new classes as handlers."""

    def __new__(mcs, name, bases, attrs):
        if name == 'NewBase':
            return super(HandlerMeta, mcs).__new__(mcs, name, bases, attrs)
        if 'options' not in attrs:
            raise RegistrationError('Class {0!r} missed "options" attribute.'.
                                    format(name))
        klass = super(HandlerMeta, mcs).__new__(mcs, name, bases, attrs)
        if klass.meta.abstract or not klass.meta.enabled:
            # Don't register abstract classes.
            return klass
        return thriftpool.register(**klass.meta)(klass)


class BaseHandler(with_metaclass(HandlerMeta)):
    """Base class for handler."""

    class options:
        abstract = True



"""Execute external commands by this worker."""
from __future__ import absolute_import

import logging

from pyuv import Pipe
from gaffer.events import EventEmitter

from thriftworker.utils.decorators import cached_property
from thriftworker.utils.loop import in_loop
from thriftworker.utils.mixin import LoopMixin

from thriftpool.components.base import StartStopComponent
from thriftpool.rpc.transport import Consumer
from thriftpool.utils.mixin import LogsMixin

logger = logging.getLogger(__name__)


class Stream(object):
    """Wrapper for plain file descriptor."""

    def __init__(self, loop, fd):
        self.loop = loop
        self.channel = Pipe(loop)
        self.fd = fd
        self._emitter = EventEmitter(loop)

    def start(self):
        self._emitter.subscribe("WRITE", self._on_write)
        self.channel.open(self.fd)
        self.channel.start_read(self._on_read)

    def write(self, data):
        self._emitter.publish("WRITE", data)

    def subscribe(self, listener):
        self._emitter.subscribe('READ', listener)

    def unsubscribe(self, listener):
        self._emitter.unsubscribe('READ', listener)

    def stop(self, all_events=False):
        if self.channel.active:
            self.channel.close()

        if all_events:
            self._emitter.close()

    def _on_write(self, evtype, data):
        self.channel.write(data)

    def _on_read(self, handle, data, error):
        if not data:
            return

        msg = dict(event='READ', data=data)
        self._emitter.publish('READ', msg)


class PerspectiveBroker(LogsMixin, LoopMixin):
    """Execute commands provided through pipe."""

    def __init__(self, app, controller):
        self.app = app
        self.controller = controller
        super(PerspectiveBroker, self).__init__()

    @cached_property
    def incoming_stream(self):
        return Stream(self.loop, self.controller.incoming_fd)

    @cached_property
    def outgoing_stream(self):
        return Stream(self.loop, self.controller.outgoing_fd)

    @property
    def streams(self):
        return [self.incoming_stream,
                self.outgoing_stream]

    @cached_property
    def consumer(self):
        return Consumer(self.loop,
                        incoming=self.incoming_stream,
                        outgoing=self.outgoing_stream,
                        handler=self.controller)

    @in_loop
    def start(self):
        for stream in self.streams:
            stream.start()
        self.consumer.start()

    @in_loop
    def stop(self):
        self.consumer.stop()
        for stream in self.streams:
            stream.stop(all_events=True)


class PerspectiveBrokerComponent(StartStopComponent):

    name = 'worker.broker'
    requires = ('loop', 'acceptors', 'worker')

    def create(self, parent):
        return PerspectiveBroker(parent.app, parent)



from __future__ import absolute_import

import logging

from thriftpool.utils.structures import AggregatedView

__all__ = ['Configuration']

DEFAULT_SETTINGS = dict(
    DEBUG=True,
    DEFAULT_LOG_FMT="[%(asctime)s %(levelname)s] %(message)s",
    DEFAULT_WORKER_LOG_FMT="[%(asctime)s %(levelname)s] [%(process)d] %(message)s",
    LOGGING_LEVEL=logging.DEBUG,
    LOG_REQUESTS=False,
    LOG_TORNADO_REQUESTS=False,
    LOG_FILE=None,
    LOG_FORCE_COLORIZED=False,
    REDIRECT_STDOUT=True,
    SLOTS=[],
    PROCESS_NAME='thriftpool',
    MODULES=[],
    PROTOCOL_FACTORY_CLS='thrift.protocol.TBinaryProtocol'
                         ':TBinaryProtocolAcceleratedFactory',
    SERVICE_PORT_RANGE=(10000, 20000),
    WORKER_TYPE='sync',
    WORKERS=1,
    WORKER_TTL=None,
    CONCURRENCY=1,
    TORNADO_ENDPOINTS=[],
    #: How long we should wait for process initialization.
    PROCESS_START_TIMEOUT=60.0,
    #: How long we should wait for process termination.
    PROCESS_STOP_TIMEOUT=60.0,
)


class Configuration(AggregatedView):
    """Default application configuration."""

    def __init__(self, config):
        super(Configuration, self).__init__(config)
        self.add_default(DEFAULT_SETTINGS)



"""Contains patched processor."""
from __future__ import absolute_import

from collections import namedtuple

from thriftworker.utils.proxy import Proxy

from thriftpool import thriftpool
from thriftpool.utils.local import LocalStack


class Request(namedtuple('Request', ('handler', 'method', 'args',
                                     'kwargs', 'service_name'))):
    """Describe thrift request."""


class RequestStack(object):
    """Store thrift requests."""

    Request = Request

    def __init__(self):
        self.stack = LocalStack()

    def add(self, handler, method, args, kwargs, service_name):
        """Register new request."""
        request = self.Request(handler, method, args, kwargs, service_name)
        self.stack.push(request)
        return request

    @property
    def current(self):
        """Return current request."""
        return self.stack.top

    def to_dict(self):
        return {ident: [(request.service_name, request.method.__name__,
                         request.args, request.kwargs)
                        for request in d.get('stack', [])]
                for ident, d in self.stack}

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.stack.pop()


#: Get current thrift request.
current_request = Proxy(lambda: thriftpool.request_stack.current)



from __future__ import absolute_import

from logging import getLogger

from thriftpool.components.base import Namespace
from thriftpool.controllers.base import Controller

logger = getLogger(__name__)


class ManagerNamespace(Namespace):

    name = 'manager'

    def modules(self):
        return ['thriftpool.components.loop',
                'thriftpool.components.listeners',
                'thriftpool.components.processes',
                'thriftpool.components.watcher']


class ManagerController(Controller):

    Namespace = ManagerNamespace

    listeners = None
    processes = None



from thriftpool.remote.ThriftPool import Iface
from thriftpool.remote.ThriftPool import Processor
from thriftpool.base import BaseHandler


class Handler(BaseHandler, Iface):

    class options:
        name = 'ThriftPool'
        processor = Processor

    def ping(self):
        pass

    def echoString(self, s):
        return s



from __future__ import absolute_import

from six import StringIO

from thriftpool.tests.utils import TestCase
from thriftpool.utils.structures import AttributeDict, AggregatedView, \
    DependencyGraph


class TestDependencyGraph(TestCase):

    def graph1(self):
        return DependencyGraph([
            ('A', []),
            ('B', []),
            ('C', ['A']),
            ('D', ['C', 'B']),
        ])

    def test_repr(self):
        self.assertTrue(repr(self.graph1()))

    def test_topsort(self):
        order = self.graph1().topsort()
        # C must start before D
        self.assertLess(order.index('C'), order.index('D'))
        # and B must start before D
        self.assertLess(order.index('B'), order.index('D'))
        # and A must start before C
        self.assertLess(order.index('A'), order.index('C'))

    def test_edges(self):
        self.assertItemsEqual(
            list(self.graph1().edges()),
            ['C', 'D'],
        )

    def test_items(self):
        self.assertDictEqual(
            dict(self.graph1().items()),
            {'A': [], 'B': [], 'C': ['A'], 'D': ['C', 'B']},
        )

    def test_to_dot(self):
        s = StringIO()
        self.graph1().to_dot(s)
        self.assertTrue(s.getvalue())


class TestAttributeDict(TestCase):

    def test_getattr__setattr(self):
        x = AttributeDict({'foo': 'bar'})
        self.assertEqual(x['foo'], 'bar')
        with self.assertRaises(AttributeError):
            x.bar
        x.bar = 'foo'
        self.assertEqual(x['bar'], 'foo')


class TestConfigurationView(TestCase):

    def setUp(self):
        view = self.view = AggregatedView({'changed_key': 1, 'both': 2})
        view.add_default({'default_key': 1, 'both': 1})

    def test_setdefault(self):
        self.assertEqual(self.view.setdefault('both', 36), 2)
        self.assertEqual(self.view.setdefault('new', 36), 36)

    def test_get(self):
        self.assertEqual(self.view.get('both'), 2)
        sp = object()
        self.assertIs(self.view.get('nonexisting', sp), sp)

    def test_update(self):
        changes = dict(self.view._changes)
        self.view.update(a=1, b=2, c=3)
        self.assertDictEqual(self.view._changes,
                             dict(changes, a=1, b=2, c=3))

    def test_contains(self):
        self.assertIn('changed_key', self.view)
        self.assertIn('default_key', self.view)
        self.assertNotIn('new', self.view)

    def test_repr(self):
        self.assertIn('changed_key', repr(self.view))
        self.assertIn('default_key', repr(self.view))

    def test_iter(self):
        expected = {'changed_key': 1,
                    'default_key': 1,
                    'both': 2}
        self.assertDictEqual(dict(self.view.items()), expected)
        self.assertItemsEqual(list(iter(self.view)),
                              list(expected.keys()))
        self.assertItemsEqual(list(self.view.keys()), list(expected.keys()))
        self.assertItemsEqual(list(self.view.values()),
                              list(expected.values()))



# -*- coding: utf-8 -*-
"""Implements controllers.

This file was copied and adapted from celery.

:copyright: (c) 2009 - 2012 by Ask Solem.
:license: BSD, see LICENSE for more details.

"""
from __future__ import absolute_import

import uuid

from logging import getLogger

from thriftworker.utils.imports import qualname
from thriftworker.utils.finalize import Finalize
from thriftworker.utils.decorators import cached_property

from thriftpool.components.base import Namespace
from thriftpool.exceptions import SystemTerminate
from thriftpool.utils.mixin import LogsMixin
from thriftpool.utils.signals import signals

__all__ = ['Controller']

logger = getLogger(__name__)


class Controller(LogsMixin):

    app = None

    wait_for_shutdown = True
    register_signals = True
    ignore_interrupt = False

    RUNNING = 0x1
    CLOSED = 0x2
    TERMINATED = 0x3

    Namespace = Namespace

    def __init__(self):
        self._running = 0
        self._state = None
        self._finalize = Finalize(self, self.stop, exitpriority=1)
        self.ident = uuid.uuid4()
        self.on_before_init()
        self.components = []
        self.namespace = self.Namespace(app=self.app).apply(self)

    @cached_property
    def _shutdown_complete(self):
        return self.app.env.Event()

    def register_signal_handler(self):
        self._debug('Register signal handlers')
        if not self.ignore_interrupt:
            signals['SIGINT'] = lambda signum, frame: self.stop()
        else:
            signals['SIGINT'] = lambda signum, frame: None
        signals['SIGTERM'] = lambda signum, frame: self.stop()
        signals['SIGQUIT'] = lambda signum, frame: self.terminate()

    def on_before_init(self):
        self.app.loader.on_before_init(self)
        self.app.finalize()

    def on_start(self):
        self.app.loader.on_start()

    def after_start(self):
        self.app.loader.after_start()

    def on_shutdown(self):
        self.app.loader.on_shutdown()

    def start(self):
        self._state = self.RUNNING

        if self.register_signals:
            self.register_signal_handler()

        self.on_start()

        try:
            for component in self.components:
                self._debug('Starting %s...', qualname(component))
                if component is not None:
                    component.start()
                self._running += 1
                self._debug('%s OK!', qualname(component))
        except SystemTerminate as exc:
            self._debug('Terminating server: %r', exc, exc_info=True)
            self.terminate()
        except Exception as exc:
            self._error('Unrecoverable error: %r', exc, exc_info=True)
            self.terminate()
        except (KeyboardInterrupt, SystemExit):
            self._debug('Terminating from keyboard')
            self.stop()

        self._debug('Whole controller started!')

        self.after_start()

        # we can't simply execute Event.wait because signal processing will
        # not work in this case
        while self.wait_for_shutdown and not self._shutdown_complete.is_set():
            self._shutdown_complete.wait(1e100)

    def _shutdown(self, warm=True):
        what = 'Stopping' if warm else 'Terminating'

        if self._state in (self.CLOSED, self.TERMINATED):
            return

        if self._state != self.RUNNING or \
                self._running != len(self.components):
            # Not fully started, can safely exit.
            self._state = self.TERMINATED
            self._shutdown_complete.set()
            return

        self._state = self.CLOSED

        for component in reversed(self.components):
            self._debug('%s %s...', what, qualname(component))
            if component:
                stop = component.stop
                if not warm:
                    stop = getattr(component, 'terminate', None) or stop
                stop()

        self.on_shutdown()

        self._debug('Whole controller stopped!')

        self._state = self.TERMINATED
        self._shutdown_complete.set()

    @property
    def is_running(self):
        return self._state == self.RUNNING

    def stop(self):
        """Graceful shutdown of the worker server."""
        self._info('Try to stop controller!')
        self._shutdown(warm=True)

    def terminate(self):
        """Not so graceful shutdown of the worker server."""
        self._info('Try to terminate controller!')
        self._shutdown(warm=False)



from __future__ import absolute_import

from thriftworker.utils.loop import in_loop

from thriftpool.app import current_app

from .transport import Producer
from .client import Client


class Clients(object):
    """Manage clients."""

    Producer = Producer
    Client = Client

    def __init__(self):
        self._producers = {}
        self._clients = {}
        super(Clients, self).__init__()

    def __iter__(self):
        return iter(self._clients)

    def __getitem__(self, key):
        return self._clients[key]

    def __contains__(self, key):
        return key in self._clients

    def keys(self):
        return self._clients.keys()

    def register(self, process, callback=None, **kwargs):
        """Create new producer for given process."""
        incoming = process.streams['incoming']
        outgoing = process.streams['outgoing']
        producer = self._producers[process.id] = \
            self.Producer(current_app.loop, incoming, outgoing)
        producer.start()
        client = self._clients[process.id] = self.Client(producer)
        if callback is not None:
            client.spawn(callback, process=process, **kwargs)

    def unregister(self, pid):
        """Stop and remove producer by process id."""
        self._clients.pop(pid, None)
        producer = self._producers.pop(pid, None)
        if producer is not None:
            producer.stop()

    def clear(self):
        """Unregister all clients."""
        for pid in list(self._producers):
            self.unregister(pid)

    @in_loop
    def spawn(self, run, *args, **kwargs):
        """Map given function to all clients."""
        greenlets = {}
        for key, client in self._clients.items():
            greenlets[key] = client.spawn(run, *args, **kwargs)
        return greenlets



from __future__ import absolute_import

import sys
import os
import argparse
import traceback

import thriftpool
from thriftpool.utils.platforms import EX_FAILURE, EX_OK
from thriftpool.utils.term import colored, isatty


class Option(object):
    """Describe an option for parser."""

    def __init__(self, *args, **kwargs):
        self.args = args
        self.kwargs = kwargs

    def apply(self, parser):
        parser.add_argument(*self.args, **self.kwargs)


class Error(Exception):

    def __init__(self, reason, status=EX_FAILURE):
        self.reason = reason
        self.status = status
        super(Error, self).__init__(reason, status)

    def __str__(self):
        return self.reason


class BaseCommand(object):
    """Abstract command. Other commands should be subclass of this."""

    #: Specify parser class.
    Parser = argparse.ArgumentParser

    #: Specify options for this command.
    options = ()

    #: Argument list used in help.
    args = ''

    #: Application version.
    version = thriftpool.__version__

    def __init__(self, app=None, stdout=sys.stdout, stderr=sys.stderr):
        self.app = thriftpool.thriftpool if app is None else app
        self.colored = colored(enabled=all([isatty(out) for out
                                            in (sys.stderr, sys.stdout)]))
        self.stdout = stdout
        self.stderr = stderr

    def out(self, msg, fh=None):
        """Print some output to text."""
        (fh or self.stdout).write(msg + '\n')

    def error(self, msg):
        """Print given error."""
        self.out(self.colored.red(msg), fh=self.stderr)

    def die(self, msg, status=EX_FAILURE):
        """Print message and exit with given status."""
        self.error(msg)
        sys.exit(status)

    def run(self, *args, **options):
        """This is the body of the command called by :meth:`handle_argv`."""
        raise NotImplementedError('subclass responsibility')

    def __call__(self, *args, **kwargs):
        """Dispatch given call."""
        try:
            ret = self.run(*args, **kwargs)
        except Error as exc:
            self.error(self.colored.red('Error: {0!r}'.format(exc)))
            return exc.status
        except Exception as exc:
            self.error(self.colored.red('Unhandled exception: {0!r}'
                                        .format(exc)))
            traceback.print_tb(sys.exc_info()[2], file=self.stderr)
            return EX_FAILURE

        return ret if ret is not None else EX_OK

    @property
    def description(self):
        """Shortcut for class doc-string."""
        return self.__doc__

    @property
    def usage(self):
        """Returns the command-line usage string for this app."""
        return '%(prog)s [options] {0.args}'.format(self)

    def parser_options(self):
        """Additional parser options."""
        return {}

    def create_parser(self, prog_name):
        """Create new parser."""
        return self.Parser(description=self.description,
                           usage=self.usage,
                           prog=prog_name,
                           **self.parser_options())

    def prepare_parser(self, parser):
        """Prepare parser for work."""
        parser.add_argument('-v', '--version', action='version',
                            version=self.version)
        for option in self.options:
            option.apply(parser)
        return parser

    def parse_options(self, prog_name, arguments):
        """Parse the available options."""
        parser = self.create_parser(prog_name)
        parser = self.prepare_parser(parser)
        return parser.parse_args(arguments)

    def handle_argv(self, prog_name, argv):
        """Parses command line arguments from ``argv`` and dispatches
        to :meth:`__call__`.

        """
        args = self.parse_options(prog_name, argv)
        options = vars(args)
        return self(*argv, **options)

    def execute(self, argv=None):
        """Execute application from command line."""
        if argv is None:
            argv = list(sys.argv)

        prog_name = os.path.basename(argv[0])
        try:
            sys.exit(self.handle_argv(prog_name, argv[1:]))
        except KeyboardInterrupt:
            sys.exit(EX_FAILURE)



"""Contains component that hold listener pool."""
from __future__ import absolute_import

import logging

from thriftpool.components.base import StartStopComponent
from thriftpool.utils.mixin import LogsMixin
from thriftpool.signals import (listener_started, listener_stopped,
                                listeners_stopped, listeners_started)

logger = logging.getLogger(__name__)


class ListenersComponent(StartStopComponent):

    name = 'manager.listeners'

    def create(self, parent):
        listeners = parent.listeners = parent.app.thriftworker.listeners
        for slot in parent.app.slots:
            name, host, port, backlog = slot.name, slot.listener.host, \
                slot.listener.port, slot.listener.backlog
            listeners.register(name, host, port, backlog)


class ListenersManager(LogsMixin):

    def __init__(self, app, listeners, processes):
        self.app = app
        self.listeners = listeners
        self.processes = processes
        super(ListenersManager, self).__init__()

    def start(self):
        """Start all registered listeners."""
        slots = self.app.slots
        clients = self.processes.clients
        for listener in self.listeners:
            slot = slots[listener.name]
            listener.start()
            listener_started.send(self, listener=listener, slot=slot,
                                  app=self.app)
            clients.spawn(lambda proxy: proxy.start_acceptor(listener.name))
            self._info("Starting listener on '%s:%d' for service '%s'.",
                       listener.host, listener.port, listener.name)
        listeners_started.send(self, app=self.app)

    def stop(self):
        """Stop all registered listeners."""
        slots = self.app.slots
        clients = self.processes.clients
        for listener in self.listeners:
            slot = slots[listener.name]
            self._info("Stopping listening on '%s:%d', service '%s'.",
                       listener.host, listener.port, listener.name)
            listener_stopped.send(self, listener=listener, slot=slot,
                                  app=self.app)
            clients.spawn(lambda proxy: proxy.stop_acceptor(listener.name))
            listener.stop()
        listeners_stopped.send(self, app=self.app)


class ListenersManagerComponent(StartStopComponent):

    name = 'manager.listeners_manager'
    requires = ('loop', 'listeners', 'process_manager')

    def create(self, parent):
        return ListenersManager(parent.app, parent.listeners,
                                parent.processes)



"""Contains patched processor."""
from __future__ import absolute_import

from thrift.Thrift import TApplicationException, TMessageType


class ProcessorMixin(object):
    """Process application error if there is one."""

    def process(self, iprot, oprot):
        name, type, seqid = iprot.readMessageBegin()

        try:
            try:
                fn = self._processMap[name]
            except KeyError:
                msg = 'Unknown function %s' % (name)
                code = TApplicationException.UNKNOWN_METHOD
                raise TApplicationException(code, msg)
            else:
                fn(self, seqid, iprot, oprot)

        except TApplicationException as exc:
            oprot.writeMessageBegin(name, TMessageType.EXCEPTION, seqid)
            exc.write(oprot)
            oprot.writeMessageEnd()
            oprot.trans.flush()

        return name



from __future__ import absolute_import

import logging
from signal import SIGTERM

from pyuv import Timer

from thriftworker.utils.loop import in_loop
from thriftworker.utils.mixin import LoopMixin
from thriftworker.utils.decorators import cached_property

from thriftpool.utils.mixin import LogsMixin

from .base import StartStopComponent

logger = logging.getLogger(__name__)


class Watcher(LogsMixin, LoopMixin):

    #: How often (in seconds) we should check for process lifetime?
    resolution = 1.0

    #: Minimum repeat delay.
    repeat_delay = 60.0

    def __init__(self, app, processes):
        self.app = app
        self.processes = processes
        super(Watcher, self).__init__()

    def _loop_cb(self, handle):
        if not self.processes.initialized:
            return
        self._timer.repeat = self.resolution
        processes = self.processes
        now = self.loop.now()
        ttl = self.app.config.WORKER_TTL
        for process_id in sorted(processes):
            lifetime = (now - processes.get_start_time(process_id)) // 1000
            if ttl < lifetime:
                self._info('Send SIGTERM to %d...', process_id)
                self.app.gaffer_manager.send_signal(process_id, SIGTERM)
                self._timer.repeat = self.repeat_delay
                break

    @cached_property
    def _timer(self):
        return Timer(self.loop)

    @in_loop
    def start(self):
        if self.app.config.WORKER_TTL is None:
            return
        self._timer.start(self._loop_cb, self.repeat_delay, self.resolution)

    @in_loop
    def stop(self):
        if not self._timer.closed:
            self._timer.close()


class WatcherComponent(StartStopComponent):

    name = 'manager.watcher'
    requires = ('loop', 'process_manager')

    def create(self, parent):
        return Watcher(parent.app, parent.processes)



"""Manage process pool."""
from __future__ import absolute_import

import json
import logging
import sys
import os

from gaffer.http.util import CorsHandler
from gaffer.http_handler import HttpHandler, HttpEndpoint
from pyuv import Pipe
from six import iteritems

from thriftworker.utils.loop import in_loop
from thriftworker.utils.decorators import cached_property
from thriftworker.utils.mixin import LoopMixin

from thriftpool.app import current_app
from thriftpool.exceptions import SystemTerminate
from thriftpool.utils.mixin import LogsMixin
from thriftpool.utils.serializers import StreamSerializer
from thriftpool.rpc.manager import Clients

from .base import StartStopComponent

logger = logging.getLogger(__name__)


class BaseHandler(CorsHandler):

    def initialize(self, clients):
        self._clients = clients

    def _execute(self, *args, **kwargs):
        execute = super(BaseHandler, self)._execute
        current_app.hub.spawn(execute, *args, **kwargs)


class ClientsHandler(BaseHandler):

    def get(self):
        self.preflight()
        self.set_status(200)
        self.write(json.dumps(self._clients.keys()))


class BaseClientHandler(BaseHandler):

    def get_data(self, proxy):
        raise NotImplementedError('subclass responsibility')

    def get(self, *args):
        self.preflight()

        try:
            pid = int(args[0])
        except ValueError:
            self.set_status(400)
            self.write({"error": "bad_value"})
            return

        if pid in self._clients:
            self.set_status(200)
        else:
            self.set_status(404)
            return

        client = self._clients[pid]
        data = client.spawn(self.get_data).get()
        self.write(json.dumps(data))


class CounterHandler(BaseClientHandler):

    def get_data(self, proxy):
        return {'{0}.{1}'.format(service, method): value
                for (service, method), value in proxy.get_counters().items()}


class TimerHandler(BaseClientHandler):

    def get_data(self, proxy):
        return {'{0}.{1}'.format(service, method): value
                for (service, method), value in proxy.get_timers().items()}


class StackHandler(BaseClientHandler):

    def get_data(self, proxy):
        return {ident: [{'method': '{0}.{1}'.format(service, method),
                         'args': repr(args), 'kwargs': repr(kwargs)}
                        for (service, method, args, kwargs) in l]
                for ident, l in proxy.get_stack().items()}


class RedirectStream(object):
    """Try to write to stream asynchronous."""

    def __init__(self, loop, stream):
        self.fd = None
        self.stream = stream
        try:
            fd = self.fd = stream.fileno()
        except AttributeError:
            self.channel = None
        else:
            channel = self.channel = Pipe(loop)
            setattr(channel, 'bypass', True)
            channel.open(fd)

    def write(self, data):
        """Write data in asynchronous way."""
        if self.channel is not None and not self.channel.closed:
            self.channel.write(data)
        else:
            self.stream.write(data)


class ProcessManager(LogsMixin, LoopMixin):
    """Start and manage workers."""

    process_name = 'worker'
    name_template = '[thriftworker-{0}]' \
                    ' -c {1.CONCURRENCY}' \
                    ' -k {1.WORKER_TYPE}'
    gevent_monkey = 'from gevent.monkey import patch_all; patch_all();'
    script = 'from thriftpool.bin.thriftworker import main; main();'

    def __init__(self, app, listeners, controller):
        self.clients = Clients()
        self.serializers = StreamSerializer()
        self.app = app
        self.listeners = listeners
        self.controller = controller
        # store process id and start time
        self._bootstrapped = {}
        super(ProcessManager, self).__init__()

    def __iter__(self):
        return iter(self._bootstrapped)

    def get_start_time(self, process_id):
        """When process was registered?"""
        return self._bootstrapped.get(process_id)

    @property
    def manager(self):
        return self.app.gaffer_manager

    @cached_property
    def _is_started(self):
        return self.app.env.RealEvent()

    @cached_property
    def _is_stopped(self):
        return self.app.env.RealEvent()

    @cached_property
    def _stdout(self):
        return RedirectStream(self.loop, sys.stdout)

    @cached_property
    def _stderr(self):
        return RedirectStream(self.loop, sys.stderr)

    @property
    def initialized(self):
        """All workers started or not?"""
        state = self.manager.get_process(self.process_name)
        return len(self._bootstrapped) >= state.numprocesses

    def _bootstrap_process(self, proxy, process):
        # Change name of process.
        name = self.name_template.format(process.id, self.app.config)
        proxy.change_title(name)

        # Register acceptors in remote process.
        proxy.register_acceptors({i: listener.name
            for i, listener in iteritems(self.listeners.enumerated)})

        for listener in self.listeners:
            if listener.started:
                proxy.start_acceptor(listener.name)

        # Notify about process initialization.
        self._bootstrapped[process.id] = self.loop.now()
        self._info('Worker %d initialized.', process.id)
        if self.initialized:
            self._info('Workers initialization done.')
            self._is_started.set()

    def _do_handshake(self, process):
        # Pass application to created process.
        stream = process.streams['handshake']
        stream.write(self.serializers.encode_with_length(self.app))

        def handshake_done(evtype, info):
            stream.unsubscribe(handshake_done)
            # Process exited and we do the same.
            if not process.active:
                return
            self.clients.register(process, self._bootstrap_process)

        # Wait for worker answer.
        stream.subscribe(handshake_done)

    def _redirect_io(self, process):
        """Redirect stdout & stderr."""
        monitor_io = (lambda evtype, msg:
                      (evtype == 'err' and self._stderr or self._stdout)
                      .write(msg['data']))
        process.monitor_io('.', monitor_io)

    def _on_event(self, evtype, msg):
        """Handle process events."""
        if msg['name'] != self.process_name:
            # Not our process.
            return

        if evtype == 'exit':
            # Log exit event.
            pid, term_signal, exit_status = \
                msg['pid'], msg['term_signal'], msg['exit_status']
            log = (exit_status != 0 or term_signal not in (0, 15)) \
                and self._critical or self._info
            log('Worker %d exited with term signal %d and exit status %d.',
                pid, term_signal, exit_status)

        elif evtype == 'spawn':
            # Log spawn event.
            self._info('Worker %d spawned with pid %d.',
                       msg['pid'], msg['os_pid'])

        if evtype == 'spawn' and self.controller.is_running:
            # New process spawned, handle event.
            process = self.manager.get_process(msg['pid'])
            self._redirect_io(process)
            self._do_handshake(process)

        elif evtype == 'exit':
            # Process exited, handle event.
            self._bootstrapped.pop(msg['pid'])
            self.clients.unregister(msg['pid'])

    def _create_proc_kwargs(self):
        """Create arguments for worker."""
        config = self.app.config
        worker_type = config.WORKER_TYPE
        if worker_type == 'gevent':
            startup_line = '{0} {1}'.format(self.gevent_monkey, self.script)
        elif worker_type == 'sync':
            startup_line = self.script
        else:
            raise NotImplementedError()
        return dict(cmd=sys.executable,
                    args=['-c', '{0}'.format(startup_line)],
                    redirect_output=['out', 'err'],
                    custom_streams=['handshake', 'incoming', 'outgoing'],
                    custom_channels=self.listeners.channels,
                    env=dict(os.environ, IS_WORKER='1'),
                    numprocesses=config.WORKERS,
                    redirect_input=True,
                    graceful_timeout=config.PROCESS_STOP_TIMEOUT)

    def _create_apps(self):
        """Create applications for gaffer."""
        apps = []
        options = dict(clients=self.clients)
        handlers = [
            (r'/timers', ClientsHandler, options),
            (r'/timers/([0-9^/]+)', TimerHandler, options),
            (r'/counters', ClientsHandler, options),
            (r'/counters/([0-9^/]+)', CounterHandler, options),
            (r'/stack', ClientsHandler, options),
            (r'/stack/([0-9^/]+)', StackHandler, options),
        ]
        endpoints = self.app.config.TORNADO_ENDPOINTS
        if endpoints:
            apps.append(HttpHandler(handlers=handlers,
                log_function=self.app.log.log_tornado_request,
                endpoints=[HttpEndpoint(uri=uri) for uri in endpoints]))
        return apps

    @in_loop
    def _setup(self):
        manager = self.manager
        manager.subscribe('.', self._on_event)
        manager.add_process(self.process_name, **self._create_proc_kwargs())
        manager.start(apps=self._create_apps())

    def start(self):
        self._setup()
        self._is_started.wait(self.app.config.PROCESS_START_TIMEOUT)
        if not self._is_started.is_set():
            self._error('Timeout when starting processes.')
            self._teardown()
            raise SystemTerminate()

    @in_loop
    def _teardown(self):
        self.clients.clear()
        stop_callback = lambda *args: self._is_stopped.set()
        self.manager.stop(stop_callback)

    def stop(self):
        self._teardown()
        self._is_stopped.wait(self.app.config.PROCESS_STOP_TIMEOUT * 2)
        if not self._is_stopped.is_set():
            self._error('Timeout when terminating processes.')
            raise SystemTerminate()


class ProcessManagerComponent(StartStopComponent):

    name = 'manager.process_manager'
    requires = ('loop', 'listeners')

    def create(self, parent):
        processes = parent.processes = \
            ProcessManager(parent.app, parent.listeners, parent)
        return processes



from __future__ import absolute_import

import argparse

from six import with_metaclass, iteritems

from thriftworker.utils.decorators import cached_property

from thriftpool.bin.base import BaseCommand
from thriftpool.bin.thriftpoold import ManagerCommand
from thriftpool.utils.mixin import SubclassMixin


def indent(t, indent=0):
    """Indent text."""
    return '\n'.join(' ' * indent + p for p in t.split('\n'))


class Formatter(argparse.HelpFormatter):
    """Print list of commands with help."""

    umbrella = None

    def _metavar_formatter(self, action, default_metavar):
        if action.choices is None:
            return super(Formatter, self)._metavar_formatter(action,
                                                             default_metavar)

        white = self.umbrella.colored.white
        parts = []
        for name, parser in iteritems(action.choices):
            parts.append(indent('+ {0}: {1}'.format(white(name),
                                                    parser.description), 2))
        result = '\n' + '\n'.join(parts)

        return lambda tuple_size: (result, ) * tuple_size


class Parser(argparse.ArgumentParser):
    """Overwrite default parser."""


class UmbrellaCommand(BaseCommand, SubclassMixin):
    """Umbrella for all commands."""

    #: Specify new parser.
    Parser = Parser

    #: Dictionary that contains existed subcommands.
    subcommand_classes = {}

    @cached_property
    def Formatter(self):
        """Create new :class:`Formatter` bounded to this class."""
        return self.subclass_with_self(Formatter, attribute='umbrella')

    @cached_property
    def subcommands(self):
        """Initialize all subcommands."""
        return {name: cls(app=self.app) for name, cls
                in iteritems(self.subcommand_classes)}

    def parser_options(self):
        return dict(formatter_class=self.Formatter)

    def prepare_parser(self, parser):
        """Prepare parser for work."""
        parser = super(UmbrellaCommand, self).prepare_parser(parser)
        subparsers = parser.add_subparsers(dest='subparser_name')
        for name, command in iteritems(self.subcommands):
            subparser = subparsers.add_parser(name,
                                              description=command.description,
                                              usage=command.usage)
            command.prepare_parser(subparser)
        return parser

    def run(self, *args, **options):
        subparser_name = options.pop('subparser_name')
        command = self.subcommands[subparser_name]
        return command.run(*args[1:], **options)


class SubCommandMeta(type):
    """Metaclass that register sub-commands."""

    def __new__(cls, name, bases, attributes):
        is_abstract = attributes['abstract'] = \
            attributes.get('abstract', False)
        klass = (super(SubCommandMeta, cls)
                 .__new__(cls, name, bases, attributes))
        if not is_abstract:
            UmbrellaCommand.subcommand_classes[klass.__name__.lower()] = klass
        return klass


class abstract(with_metaclass(SubCommandMeta, BaseCommand)):
    """Base class for sub-commands."""

    #: Register this command in umbrella?
    abstract = True


class list_slots(abstract):
    """Print all registered slots."""

    def format_slot(self, slot):
        white = self.colored.white
        yellow = self.colored.yellow
        parts = ['+ {0}: '.format(white(slot.name))]
        parts.extend([indent(s, 4) for s in (
            'listen: {0}:{1}'.format(yellow(slot.listener.host),
                                     yellow(slot.listener.port or 0)),
            'backlog: {0}'.format(yellow(slot.listener.backlog)),
            'processor: {0}'.format(yellow(slot.service.Processor)),
            'handler: {0}'.format(yellow(slot.service.Handler)),
        )])
        return '\n'.join(parts).strip()

    def run(self, *args, **options):
        self.app.finalize()
        self.out('Registered slots:\n')
        for slot in self.app.slots:
            self.out(self.format_slot(slot) + '\n')


class manager(abstract, ManagerCommand):
    """Run manager daemon. Same as `thriftpoold`."""


def main():
    UmbrellaCommand().execute()


if __name__ == '__main__':
    main()



from __future__ import absolute_import

import signal as _signal

from six import iteritems, string_types

__all__ = ['signals']


class Signals(object):
    """Convenience interface to :mod:`signals`.

    If the requested signal is not supported on the current platform,
    the operation will be ignored.

    **Examples**:

    .. code-block:: python

        >>> from celery.platforms import signals

        >>> signals['INT'] = my_handler

        >>> signals['INT']
        my_handler

        >>> signals.supported('INT')
        True

        >>> signals.signum('INT')
        2

        >>> signals.ignore('USR1')
        >>> signals['USR1'] == signals.ignored
        True

        >>> signals.reset('USR1')
        >>> signals['USR1'] == signals.default
        True

        >>> signals.update(INT=exit_handler,
        ...                TERM=exit_handler,
        ...                HUP=hup_handler)

    """

    ignored = _signal.SIG_IGN
    default = _signal.SIG_DFL

    def supported(self, signal_name):
        """Returns true value if ``signal_name`` exists on this platform."""
        try:
            return self.signum(signal_name)
        except AttributeError:
            pass

    def signum(self, signal_name):
        """Get signal number from signal name."""
        if isinstance(signal_name, int):
            return signal_name
        if not isinstance(signal_name, string_types) \
                or not signal_name.isupper():
            raise TypeError('signal name must be uppercase string.')
        if not signal_name.startswith('SIG'):
            signal_name = 'SIG' + signal_name
        return getattr(_signal, signal_name)

    def reset(self, *signal_names):
        """Reset signals to the default signal handler.

        Does nothing if the platform doesn't support signals,
        or the specified signal in particular.

        """
        self.update((sig, self.default) for sig in signal_names)

    def ignore(self, *signal_names):
        """Ignore signal using :const:`SIG_IGN`.

        Does nothing if the platform doesn't support signals,
        or the specified signal in particular.

        """
        self.update((sig, self.ignored) for sig in signal_names)

    def __getitem__(self, signal_name):
        return _signal.getsignal(self.signum(signal_name))

    def __setitem__(self, signal_name, handler):
        """Install signal handler.

        Does nothing if the current platform doesn't support signals,
        or the specified signal in particular.

        """
        try:
            _signal.signal(self.signum(signal_name), handler)
        except (AttributeError, ValueError):
            pass

    def update(self, _d_=None, **sigmap):
        """Set signal handlers from a mapping."""
        for signal_name, handler in iteritems(dict(_d_ or {}, **sigmap)):
            self[signal_name] = handler

signals = Signals()



# -*- encoding: utf-8 -*-
from __future__ import absolute_import

from thriftpool.tests.utils import TestCase
from thriftpool.utils.encoding import smart_str, smart_unicode


class TestEncoding(TestCase):

    def setUp(self):
        super(TestEncoding, self).setUp()
        u_str = self.unicode_str = u'строка'
        self.encoded_unicode_str = u_str.encode('utf-8')
        b_str = self.bytes_str = 'строка'
        self.decoded_bytes_str = b_str.decode('utf-8')

    def test_smart_str(self):
        self.assertEqual(self.bytes_str, smart_str(self.decoded_bytes_str))
        self.assertIsInstance(smart_str(self.decoded_bytes_str), str)
        self.assertEqual(self.bytes_str, smart_str(self.bytes_str))
        self.assertIsInstance(smart_str(self.bytes_str), str)

    def test_smart_unicode(self):
        self.assertEqual(self.unicode_str,
                         smart_unicode(self.encoded_unicode_str))
        self.assertIsInstance(smart_unicode(self.encoded_unicode_str), unicode)
        self.assertEqual(self.unicode_str, smart_unicode(self.unicode_str))
        self.assertIsInstance(smart_unicode(self.unicode_str), unicode)



from __future__ import absolute_import

import atexit
import warnings
import os

from thriftpool.utils.platforms import create_pidlock, daemonize
from thriftpool.signals import collect_excluded_fds


class Daemon(object):
    """Daemon behavior."""

    app = None

    def __init__(self, controller, pidfile=None, daemonize=False,
                 foreground=False, args=None):
        self.controller = controller
        self.pidfile = pidfile
        self.pidlock = None
        self.daemonize = daemonize
        self.foreground = foreground
        self.daemon = None
        self.args = list(args or [])
        atexit.register(self.on_exit)

    def on_start(self):
        if getattr(os, 'getuid', None) and os.getuid() == 0:
            warnings.warn(RuntimeWarning(
                'Running thriftpoold with superuser'
                ' privileges is discouraged!'))

    def start(self):
        """Start new worker."""
        self.on_start()
        if self.pidfile is not None:
            self.pidlock = create_pidlock(self.pidfile)
        if self.daemonize:
            receivers = collect_excluded_fds.send(sender=self)
            excluded = [receiver[1] for receiver in receivers if receiver[1]]
            excluded = [item for sublist in excluded for item in sublist]
            self.daemon = daemonize(self.foreground, excluded_fds=excluded)
        self.controller.start()

    def on_exit(self):
        pass



from __future__ import absolute_import

import os
import errno
import atexit
import sys
import resource
from contextlib import contextmanager

try:
    import setproctitle
except ImportError:
    setproctitle_exists = False
else:
    setproctitle_exists = True


EX_OK = getattr(os, 'EX_OK', 0)
EX_FAILURE = 1
EX_UNAVAILABLE = getattr(os, 'EX_UNAVAILABLE', 69)
EX_USAGE = getattr(os, 'EX_USAGE', 64)

DAEMON_UMASK = 0
DAEMON_WORKDIR = '/'


def set_process_title(name):
    """Change process title."""
    if not setproctitle_exists:
        return
    setproctitle.setproctitle(name)


def fileno(f):
    try:
        return f.fileno()
    except AttributeError:
        return f


def get_fdmax(default=None):
    """Returns the maximum number of open file descriptors
    on this system.

    :keyword default: Value returned if there's no file
                      descriptor limit.

    """
    fdmax = resource.getrlimit(resource.RLIMIT_NOFILE)[1]
    if fdmax == resource.RLIM_INFINITY:
        return default
    return fdmax


@contextmanager
def ignore_EBADF():
    try:
        yield
    except OSError as exc:
        if exc.errno != errno.EBADF:
            raise


class LockFailed(Exception):
    pass


class PIDLock(object):
    """Create PID lock and work with it as with resource."""

    FLAGS = os.O_CREAT | os.O_EXCL | os.O_WRONLY
    MODE = ((os.R_OK | os.W_OK) << 6) | ((os.R_OK) << 3) | ((os.R_OK))

    def __init__(self, path):
        self.path = os.path.abspath(path)

    def read(self):
        """Reads and returns the writed PID lock."""
        try:
            fh = open(self.path, 'r')
        except IOError as exc:
            if exc.errno == errno.ENOENT:
                return
            raise

        try:
            line = fh.readline()
            if line.strip() == line:  # must contain '\n'
                raise ValueError(
                    'Partial or invalid PID lock {0.path}'.format(self))
        finally:
            fh.close()

        try:
            return int(line.strip())
        except ValueError:
            raise ValueError('PID lock {0.path} invalid.'.format(self))

    def write(self, pid):
        """Write given PID to lock."""
        fd = os.open(self.path, self.FLAGS, self.MODE)
        fh = os.fdopen(fd, 'w')
        try:
            fh.write('{0}\n'.format(pid))
            # flush and sync so that the re-read below works.
            fh.flush()
            try:
                os.fsync(fd)
            except AttributeError:
                pass
        finally:
            fh.close()

        # Check that PID wasn't change.
        if self.read() != pid:
            raise LockFailed('PID lock changed!')

    def remove(self):
        """Remove PID lock if exists."""
        try:
            os.unlink(self.path)
        except OSError as exc:
            if exc.errno == errno.ENOENT:
                return
            raise

    def exists(self):
        """Returns True if the PID lock exists."""
        return os.path.exists(self.path)

    @staticmethod
    def process_exists(pid):
        """Check that process exists."""
        try:
            os.kill(pid, 0)
        except os.error as exc:
            if exc.errno == errno.ESRCH:
                return False
        return True

    def maybe_remove(self):
        """Remove existed PID lock if we can."""
        if not self.exists():
            # No PID exists, return.
            return

        pid = self.read()
        if pid is None:
            # No PID found.
            return

        if self.process_exists(pid):
            raise LockFailed('PID lock exists.')

        self.remove()

    def acquire(self):
        """Try to write PID lock."""
        self.maybe_remove()
        self.write(os.getpid())

    def release(self, *args):
        """Try to remove PID lock."""
        self.remove()


def create_pidlock(pidfile):
    """Create PID lock, exit if fail."""
    pid = PIDLock(pidfile)
    try:
        pid.acquire()
    except LockFailed:
        raise SystemExit("Error: PID file ({0}) exists.".format(pidfile))
    atexit.register(pid.release)
    return pid


class Daemon(object):
    """Fork and close needed resources."""

    _is_open = False
    workdir = DAEMON_WORKDIR
    umask = DAEMON_UMASK

    def __init__(self, fake=False, excluded_fds=None):
        self.fake = fake
        self.standart_fds = (sys.stdin, sys.stdout, sys.stderr)
        self.excluded_fds = tuple(excluded_fds or [])

    @property
    def preserved_fds(self):
        return self.standart_fds + self.excluded_fds

    def redirect_to_null(self, fd):
        if fd:
            dest = os.open(os.devnull, os.O_RDWR)
            os.dup2(dest, fd)

    def _detach(self):
        if os.fork() == 0:      # first child
            os.setsid()         # create new session
            if os.fork() > 0:   # second child
                os._exit(0)
        else:
            os._exit(0)
        return self

    def open(self):
        if not self._is_open:
            if not self.fake:
                self._detach()

            os.chdir(self.workdir)
            os.umask(self.umask)

            preserve = [fileno(f) for f in self.preserved_fds if fileno(f)]
            for fd in reversed(range(get_fdmax(default=2048))):
                if fd not in preserve:
                    with ignore_EBADF():
                        os.close(fd)

            for fd in self.standart_fds:
                self.redirect_to_null(fileno(fd))

            self._is_open = True

    def close(self, *args):
        if self._is_open:
            self._is_open = False


def daemonize(fake=False, excluded_fds=None):
    """Try to start daemon."""
    daemon = Daemon(fake, excluded_fds)
    daemon.open()
    atexit.register(daemon.close)
    return daemon



#
# Autogenerated by Thrift Compiler (0.9.0)
#
# DO NOT EDIT UNLESS YOU ARE SURE THAT YOU KNOW WHAT YOU ARE DOING
#
#  options string: py:new_style,utf8strings,slots,dynamic
#

from thrift.Thrift import TType, TMessageType, TException, TApplicationException
from ttypes import *




# -*- coding: utf-8 -*-
from __future__ import absolute_import

from mock import Mock
from thrift.Thrift import TException, TApplicationException

from thriftpool.tests.utils import TestCase
from thriftpool.exceptions import WrappingError
from thriftpool.remote.ThriftPool import Iface
from thriftpool.request.handler import guarded_method, BaseWrappedHandler, \
    WrappedHandlerMeta


class UnknownException(Exception):
    pass


class NotFound(TException):
    pass


class Handler(Iface):

    def ping(self):
        pass


class WrongHandler(object):

    def ping(self):
        pass


class TestGuardedMethod(TestCase):

    def setUp(self):
        methods = self.methods = {'ping', 'fail', 'not_found'}
        attrs = {method: guarded_method(method) for method in methods}
        attrs['_wrapped_methods'] = methods
        handler = self.handler = Mock()
        handler.ping.side_effect = lambda value: value
        handler.fail.side_effect = UnknownException()
        handler.not_found.side_effect = NotFound()
        handler.unknown.side_effect = UnknownException()
        cls = type('GuardedHandler', (BaseWrappedHandler,), attrs)
        self.guarded = cls(handler)
        super(TestGuardedMethod, self).setUp()

    def test_introspection(self):
        for method in self.methods:
            self.assertIn(method, vars(self.guarded))

    def test_call(self):
        value = object()
        self.assertIs(value, self.guarded.ping(value))
        self.assertTrue(self.handler.ping.called)

    def test_unwrapped_method(self):
        # unwrapped method not touched, check this
        with self.assertRaises(UnknownException):
            self.guarded.unknown()

    def test_wrapped_method(self):
        # ensure that thrift exceptions properly re-raised
        with self.assertRaises(NotFound):
            self.guarded.not_found()
        # ensure that unknown exceptions properly wrapped
        with self.assertRaises(TApplicationException):
            self.guarded.fail()


class TestHandlerMeta(TestCase):

    def test_wrapped_class(self):
        service_name = 'ThriftPool'
        attrs = dict(_handler_cls=Handler,
                     _service_name=service_name)
        cls = WrappedHandlerMeta('GuardedHandler', (object, ), attrs)
        self.assertIs(Handler, cls._handler_cls)
        self.assertIs(service_name, cls._service_name)
        methods = set(cls._wrapped_methods)
        self.assertTrue(methods)
        self.assertTrue(methods <= set(vars(Iface).keys()))
        for method in methods:
            self.assertIsInstance(getattr(cls, method), guarded_method)

    def test_wrong_class(self):
        service_name = 'ThriftPool'
        attrs = dict(_handler_cls=WrongHandler,
                     _service_name=service_name)
        with self.assertRaises(WrappingError):
            WrappedHandlerMeta('GuardedHandler', (object, ), attrs)

    def test_missing_attributes(self):
        service_name = 'ThriftPool'
        with self.assertRaises(WrappingError):
            WrappedHandlerMeta('GuardedHandler', (object, ),
                               dict(_handler_cls=Handler))
        with self.assertRaises(WrappingError):
            WrappedHandlerMeta('GuardedHandler', (object, ),
                               dict(_service_name=service_name))



from __future__ import absolute_import

import inspect

from thriftworker.utils.decorators import cached_property
from thriftworker.utils.imports import symbol_by_name

from thriftpool.app._state import current_app


def rgetattr(obj, path):
    """Get nested attribute from object.

    :param obj: object
    :param path: path to attribute

    """
    return reduce(getattr, [obj] + path.split('.'))


def _unpickle(name, args):
    """Given an attribute name and a list of args, gets
    the attribute from the current app and calls it.

    """
    return rgetattr(current_app, name)(*args)


class SubclassMixin(object):

    def subclass_with_self(self, Class, name=None, attribute='app',
                           reverse=None, **kw):
        """Subclass an app-compatible class by setting its app attribute
        to be this app instance.

        App-compatible means that the class has a class attribute that
        provides the default app it should use, e.g.
        ``class Foo: app = None``.

        :param Class: The app-compatible class to subclass.
        :keyword name: Custom name for the target class.
        :keyword attribute: Name of the attribute holding the app,
                            default is 'app'.

        """
        Class = symbol_by_name(Class)
        reverse = reverse if reverse else Class.__name__
        has_reduce_args = getattr(Class, '__reduce_args__', None) is not None

        def __reduce__(self):
            args = self.__reduce_args__() if has_reduce_args else []
            return (_unpickle, (Class.__name__, args))

        attrs = dict({attribute: self},
                     __module__=Class.__module__,
                     __doc__=Class.__doc__,
                     __reduce__=__reduce__,
                     **kw)

        return type(name or Class.__name__, (Class,), attrs)


class LogsMixin(object):
    """Simple helper for logging."""

    @cached_property
    def _logger(self):
        module = inspect.getmodule(self.__class__)
        return getattr(module, 'logger')

    def _exception(self, exc):
        self._logger.exception(exc)

    def _critical(self, msg, *args, **kwargs):
        self._logger.critical(msg, *args, **kwargs)

    def _error(self, msg, *args, **kwargs):
        self._logger.error(msg, *args, **kwargs)

    def _info(self, msg, *args, **kwargs):
        self._logger.info(msg, *args, **kwargs)

    def _debug(self, msg, *args, **kwargs):
        self._logger.debug(msg, *args, **kwargs)



from __future__ import absolute_import

import cPickle as pickle

from thriftpool.tests.utils import TestCase


class TestApp(TestCase):

    def test_pickling(self):
        self.app.loader.preload_modules()
        self.assertIn('ThriftPool', self.app.slots)
        buf = pickle.dumps(self.app)
        app = pickle.loads(buf)
        # ensure that slots serialized and loaded properly
        self.assertIn('ThriftPool', app.slots)
        # check configuration serialization
        self.assertEqual(dict(self.app.config), dict(app.config))



"""Some useful tools for logging.

This file was copied and adapted from celery.

:copyright: (c) 2009 - 2012 by Ask Solem.
:license: BSD, see LICENSE for more details.

"""
from __future__ import absolute_import

import logging
import sys
import threading
import traceback

from .encoding import smart_str
from .term import colored


LOG_LEVELS = {
    'DEBUG': logging.DEBUG,
    'INFO': logging.INFO,
    'WARNING': logging.WARNING,
    'ERROR': logging.ERROR,
    'CRITICAL': logging.CRITICAL,
    'FATAL': logging.FATAL,
}
LEVELS = {k.lower() for k in LOG_LEVELS.keys()}


def mlevel(level):
    if level and not isinstance(level, int):
        return LOG_LEVELS[level.upper()]
    return level


class ColorFormatter(logging.Formatter):

    #: Loglevel -> Color mapping.
    COLORS = colored().names
    colors = {'DEBUG': COLORS['blue'], 'WARNING': COLORS['yellow'],
              'ERROR': COLORS['red'], 'CRITICAL': COLORS['magenta']}

    def __init__(self, fmt=None, datefmt=None, use_color=True):
        logging.Formatter.__init__(self, fmt, datefmt)
        self.use_color = use_color

    def formatException(self, ei):
        r = logging.Formatter.formatException(self, ei)
        if isinstance(r, str):
            return smart_str(r)
        return r

    def format(self, record):
        levelname = record.levelname
        color = self.colors.get(levelname)

        if self.use_color and color:
            record.msg = smart_str(color(record.msg))

        return smart_str(logging.Formatter.format(self, record))


class LoggingProxy(object):
    """Forward file object to :class:`logging.Logger` instance.

    :param logger: The :class:`logging.Logger` instance to forward to.
    :param loglevel: Loglevel to use when writing messages.

    """
    mode = 'w'
    name = None
    closed = False
    loglevel = logging.ERROR
    _thread = threading.local()

    def __init__(self, logger, loglevel=None):
        self.logger = logger
        self.loglevel = loglevel or logging.WARNING
        self._safewrap_handlers()

    def _safewrap_handlers(self):
        """Make the logger handlers dump internal errors to
        `sys.__stderr__` instead of `sys.stderr` to circumvent
        infinite loops."""

        def wrap_handler(handler):                  # pragma: no cover

            class WithSafeHandleError(logging.Handler):

                def handleError(self, record):
                    exc_info = sys.exc_info()
                    try:
                        try:
                            traceback.print_exception(exc_info[0],
                                                      exc_info[1],
                                                      exc_info[2],
                                                      None, sys.__stderr__)
                        except IOError:
                            pass    # see python issue 5971
                    finally:
                        del(exc_info)

            handler.handleError = WithSafeHandleError().handleError

        return [wrap_handler(l) for l in self.logger.handlers]

    def write(self, data):
        """Write message to logging object."""
        if getattr(self._thread, 'recurse_protection', False):
            # Logger is logging back to this file, so stop recursing.
            return
        data = data.strip('\n')
        if data.strip() and not self.closed:
            self._thread.recurse_protection = True
            try:
                self.logger.log(self.loglevel, smart_str(data))
            finally:
                self._thread.recurse_protection = False

    def writelines(self, sequence):
        """`writelines(sequence_of_strings) -> None`.

        Write the strings to the file.

        The sequence can be any iterable object producing strings.
        This is equivalent to calling :meth:`write` for each string.

        """
        for part in sequence:
            self.write(part)

    def flush(self):
        """This object is not buffered so any :meth:`flush` requests
        are ignored."""
        pass

    def close(self):
        """When the object is closed, no write requests are forwarded to
        the logging object anymore."""
        self.closed = True

    def isatty(self):
        """Always returns :const:`False`. Just here for file support."""
        return False

    def fileno(self):
        return None



"""Provide some tools to patch thrift handler."""
from __future__ import absolute_import

import logging
from functools import wraps

from thrift.Thrift import TApplicationException, TException
from thrift.protocol.TBase import TExceptionBase

from thriftpool import thriftpool
from thriftpool.signals import handler_method_guarded
from thriftpool.exceptions import WrappingError

logger = logging.getLogger(__name__)


def maybe_wraps(method):
    """Ignore wrapping exceptions."""

    def inner_wrapper(fn):
        try:
            return wraps(method)(fn)
        except AttributeError:
            return fn

    return inner_wrapper


class guarded_method(object):
    """Create guarded method for handler."""

    def __init__(self, name, doc=None):
        self.__name__ = name
        self.__doc__ = doc

    def __create_method(self, obj):
        """Create bounded method."""
        handler = obj._handler
        service_name = obj._service_name
        method = getattr(handler, self.__name__)
        stack = thriftpool.request_stack
        allowed_exceptions = (TException, TExceptionBase)

        # Apply all returned by signal decorators.
        for sender, decorator in handler_method_guarded.send(sender=handler,
                                                             fn=method):
            if decorator is not None:
                method = decorator(method)

        @maybe_wraps(method)
        def inner_method(*args, **kwargs):
            """Method that handle unknown exception correctly."""
            stack.add(handler, method, args, kwargs, service_name)
            with stack:
                try:
                    return method(*args, **kwargs)
                except allowed_exceptions:
                    raise
                except Exception as exc:
                    # Catch all exceptions here, process they here. Write
                    # application exception to thrift transport.
                    logger.exception(exc)
                    code = TApplicationException.INTERNAL_ERROR
                    msg = "{0}({1})".format(type(exc).__name__, str(exc))
                    raise TApplicationException(code, msg)

        return inner_method

    def __get__(self, obj, type=None):
        if obj is None:
            return self
        try:
            return obj.__dict__[self.__name__]
        except KeyError:
            value = obj.__dict__[self.__name__] = self.__create_method(obj)
            return value


class BaseWrappedHandler(object):
    """Abstract base for wrapped handler."""

    _handler_cls = None
    _service_name = None
    _wrapped_methods = None

    def __init__(self, handler):
        self._handler = handler
        # Force methods creation.
        for item in self._wrapped_methods:
            getattr(self, item)

    def __getattr__(self, name):
        """All unknown attribute access go to wrapped handler."""
        return getattr(self._handler, name)


class WrappedHandlerMeta(type):
    """Metaclass that create handler with decorated methods."""

    def __new__(mcs, name, bases, attrs):
        # Add base class to bases.
        if BaseWrappedHandler not in bases:
            bases = (BaseWrappedHandler,) + bases
        # Ensure that service name provided.
        if '_service_name' not in attrs:
            raise WrappingError('Missing attribute "_service_name" in'
                                ' class "{0}"'.format(name))
        # Get original handler class.
        try:
            handler_cls = attrs['_handler_cls']
        except KeyError:
            raise WrappingError('Missing attribute "_handler_cls" in'
                                ' class "{0}"'.format(name))
        # Try to find methods of service.
        included = set()
        stack = set(handler_cls.__bases__)
        while stack:
            # Collect all bases and try to find all classes with name 'Iface'.
            # Methods with collected names will be decorated.
            base = stack.pop()
            if base is object:
                continue
            stack.update(base.__bases__)
            if base.__name__ == 'Iface':
                included.update(vars(base).keys())

        # No interface found, fail.
        if not included:
            raise WrappingError('No methods to wrap found in class "{0}"'.
                                format(name))

        # Find attributes that must be decorated.
        included = included - set(vars(type).keys() + ['__weakref__'])
        attrs['_wrapped_methods'] = included

        # Decorate them!
        for attr in included:
            method = guarded_method(attr, getattr(handler_cls, attr).__doc__)
            attrs[attr] = method

        # Create new class.
        return super(WrappedHandlerMeta, mcs).__new__(mcs, name, bases, attrs)



from __future__ import absolute_import

import os
import contextlib
from functools import partial

from unittest import TestCase as BaseTestCase

from thriftpool.app import _state as state


@contextlib.contextmanager
def change_cwd(cwd):
    """Temporary change current directory."""
    current_cwd = os.getcwd()
    os.chdir(cwd)
    try:
        yield
    finally:
        os.chdir(current_cwd)


@contextlib.contextmanager
def custom_settings(app, **kwargs):
    """Temporarily change settings."""
    variables = vars(app.config)
    original_underlying_dicts = variables['_underlying_dicts']
    variables['_underlying_dicts'] = [kwargs] + original_underlying_dicts
    try:
        yield
    finally:
        variables['_underlying_dicts'] = original_underlying_dicts


class Noop(object):
    """Object that do nothing."""


class TestCase(BaseTestCase):

    def setUp(self):
        # reset current state before each run
        state.default_app = None
        app = self.app = state.get_current_app()
        self.thriftworker = self.app.thriftworker
        self.custom_settings = partial(custom_settings, app)



from __future__ import absolute_import

import sys

from thriftpool.bin.base import BaseCommand
from thriftpool.utils.serializers import StreamSerializer


class WorkerCommand(BaseCommand):
    """Start ThiftPool worker."""

    def run(self, *args, **options):
        stream_fd = sys.stderr.fileno() + 1
        app = StreamSerializer().decode_from_stream(stream_fd)
        controller = app.WorkerController(stream_fd)
        controller.start()


def main():
    WorkerCommand().execute()


if __name__ == '__main__':
    main()



"""Specify signals sent when specific events happens."""
from __future__ import absolute_import

from thriftpool.utils.dispatch import Signal


#: Called before handler will be guarded. Handler may return decorator.
handler_method_guarded = Signal(providing_args=['fn'])


#: Listener was started.
listener_started = Signal(providing_args=['listener', 'slot', 'app'])


#: All listener was started.
listeners_started = Signal(providing_args=['app'])


#: Listener was stopped.
listener_stopped = Signal(providing_args=['listener', 'slot', 'app'])


#: All listeners stopped.
listeners_stopped = Signal(providing_args=['app'])


#: Called before loggers are configured.
setup_logging = Signal(providing_args=['root', 'logfile', 'loglevel'])


#: Called after loggers are configured.
after_logging_setup = Signal(providing_args=['root', 'logfile', 'loglevel'])


#: Called on application start.
app_start = Signal(providing_args=['app'])


#: Called after application start.
after_app_start = Signal(providing_args=['app'])


#: Called after application stopped.
app_shutdown = Signal(providing_args=['app'])


#: Collect file descriptors that must not be closed.
collect_excluded_fds = Signal(providing_args=[])



# -*- coding: utf-8 -*-
"""
    celery.utils.term
    ~~~~~~~~~~~~~~~~~

    Terminals and colors.

"""
from __future__ import absolute_import, unicode_literals

import platform
from six import text_type

from thriftpool.utils.encoding import smart_str, smart_unicode


BLACK, RED, GREEN, YELLOW, BLUE, MAGENTA, CYAN, WHITE = range(8)
OP_SEQ = '\033[%dm'
RESET_SEQ = '\033[0m'
COLOR_SEQ = '\033[1;%dm'
fg = lambda s: COLOR_SEQ % s

SYSTEM = platform.system()
IS_WINDOWS = SYSTEM == 'Windows'


def isatty(stream):
    return hasattr(stream, 'isatty') and stream.isatty()


class colored(object):
    """Terminal colored text.

    Example::
        >>> c = colored(enabled=True)
        >>> print(str(c.red('the quick '), c.blue('brown ', c.bold('fox ')),
        ...       c.magenta(c.underline('jumps over')),
        ...       c.yellow(' the lazy '),
        ...       c.green('dog ')))

    """

    def __init__(self, *s, **kwargs):
        self.s = s
        self.enabled = not IS_WINDOWS and kwargs.get('enabled', True)
        self.op = kwargs.get('op', '')
        self.names = {'black': self.black,
                      'red': self.red,
                      'green': self.green,
                      'yellow': self.yellow,
                      'blue': self.blue,
                      'magenta': self.magenta,
                      'cyan': self.cyan,
                      'white': self.white}

    def _add(self, a, b):
        return text_type(a) + text_type(b)

    def _fold_no_color(self, a, b):
        try:
            A = a.no_color()
        except AttributeError:
            A = text_type(a)
        try:
            B = b.no_color()
        except AttributeError:
            B = text_type(b)
        return smart_str(A) + smart_str(B)

    def no_color(self):
        if self.s:
            return reduce(self._fold_no_color, self.s)
        return ''

    def embed(self):
        prefix = ''
        if self.enabled:
            prefix = self.op
        return smart_str(prefix) + smart_str(reduce(self._add, self.s))

    def __unicode__(self):
        suffix = ''
        if self.enabled:
            suffix = RESET_SEQ
        return smart_unicode(self.embed()) + suffix

    def __str__(self):
        return smart_str(self.__unicode__())

    def __len__(self):
        return len(str(self))

    def node(self, s, op):
        return self.__class__(enabled=self.enabled, op=op, *s)

    def black(self, *s):
        return self.node(s, fg(30 + BLACK))

    def red(self, *s):
        return self.node(s, fg(30 + RED))

    def green(self, *s):
        return self.node(s, fg(30 + GREEN))

    def yellow(self, *s):
        return self.node(s, fg(30 + YELLOW))

    def blue(self, *s):
        return self.node(s, fg(30 + BLUE))

    def magenta(self, *s):
        return self.node(s, fg(30 + MAGENTA))

    def cyan(self, *s):
        return self.node(s, fg(30 + CYAN))

    def white(self, *s):
        return self.node(s, fg(30 + WHITE))

    def __repr__(self):
        return repr(self.no_color())

    def bold(self, *s):
        return self.node(s, OP_SEQ % 1)

    def underline(self, *s):
        return self.node(s, OP_SEQ % 4)

    def blink(self, *s):
        return self.node(s, OP_SEQ % 5)

    def reverse(self, *s):
        return self.node(s, OP_SEQ % 7)

    def bright(self, *s):
        return self.node(s, OP_SEQ % 8)

    def ired(self, *s):
        return self.node(s, fg(40 + RED))

    def igreen(self, *s):
        return self.node(s, fg(40 + GREEN))

    def iyellow(self, *s):
        return self.node(s, fg(40 + YELLOW))

    def iblue(self, *s):
        return self.node(s, fg(40 + BLUE))

    def imagenta(self, *s):
        return self.node(s, fg(40 + MAGENTA))

    def icyan(self, *s):
        return self.node(s, fg(40 + CYAN))

    def iwhite(self, *s):
        return self.node(s, fg(40 + WHITE))

    def reset(self, *s):
        return self.node(s or [''], RESET_SEQ)

    def __add__(self, other):
        return text_type(self) + text_type(other)



"""Main factory for this library. Single entry point for all application."""
from __future__ import absolute_import

import inspect
from threading import RLock

from gaffer.manager import Manager

from thriftworker.utils.decorators import cached_property
from thriftworker.utils.imports import instantiate
from thriftworker.app import ThriftWorker

from thriftpool.app.config import Configuration
from thriftpool.exceptions import RegistrationError
from thriftpool.utils.mixin import SubclassMixin

from ._state import set_current_app

__all__ = ['ThriftPool']


def _unpickle_app(cls, changes, slots):
    app = cls()
    app.config.update(changes)
    app.slots = slots
    app.loader.after_unpickling()
    return app


class ThriftPool(SubclassMixin):
    """Main entry point for this application."""

    #: Default loader for this application. Must provide configuration
    #: for it.
    loader_cls = 'thriftpool.app.loader:Loader'

    #: What class should be used for logging setup.
    logging_cls = 'thriftpool.app.log:Logging'

    #: Repository for registered slots.
    repo_cls = 'thriftpool.app.slots:Repository'

    #: Manager controller class.
    manager_cls = 'thriftpool.controllers.manager:ManagerController'

    #: Worker controller class.
    worker_cls = 'thriftpool.controllers.worker:WorkerController'

    #: Specify daemonizing behavior.
    daemon_cls = 'thriftpool.app.daemon:Daemon'

    #: Store active requests here.
    request_stack_cls = 'thriftpool.request.stack:RequestStack'

    def __init__(self):
        self._finalized = False
        self._finalize_mutex = RLock()
        super(ThriftPool, self).__init__()

        # set current application as default
        set_current_app(self)

    def __reduce__(self):
        # Reduce only pickles the configuration changes,
        # so the default configuration doesn't have to be passed
        # between processes.
        return (_unpickle_app, (self.__class__, self.config._changes,
                                self.slots))

    @cached_property
    def Loader(self):
        """Default loader class."""
        return self.subclass_with_self(self.loader_cls)

    @cached_property
    def loader(self):
        return self.Loader()

    @cached_property
    def config(self):
        """Empty application configuration."""
        return Configuration(self.loader.get_config())

    @cached_property
    def Logging(self):
        """Create bounded logging initialize class from :class:`.log.Logging`.
        We will call :meth:`.log.Logging.setup` on finalization to setup
        logging.

        """
        return self.subclass_with_self(self.logging_cls)

    @cached_property
    def log(self):
        """Instantiate logging initializer from bounded class."""
        return self.Logging()

    @cached_property
    def Repository(self):
        """Create bounded slots repository from :class:`.slots:Repository`."""
        return self.subclass_with_self(self.repo_cls)

    @cached_property
    def slots(self):
        """Create repository of service slots. By default it is empty."""
        return self.Repository()

    def finalize(self):
        """Make some steps before application startup."""
        with self._finalize_mutex:
            if self._finalized:
                return
            # Setup logging for whole application.
            self.log.setup()
            # Load all needed modules.
            self.loader.preload_modules()
            # Register existed services.
            for params in self.config.SLOTS:
                self.slots.register(**params)
            self._finalized = True

    @cached_property
    def protocol_factory(self):
        """Create handler instance."""
        return instantiate(self.config.PROTOCOL_FACTORY_CLS)

    @cached_property
    def thriftworker(self):
        return ThriftWorker(port_range=self.config.SERVICE_PORT_RANGE,
                            protocol_factory=self.protocol_factory,
                            pool_size=self.config.CONCURRENCY)

    @property
    def loop(self):
        return self.thriftworker.loop

    @property
    def hub(self):
        return self.thriftworker.hub

    @property
    def env(self):
        return self.thriftworker.env

    @cached_property
    def gaffer_manager(self):
        """Create process manager."""
        return Manager(loop=self.loop)

    @cached_property
    def ManagerController(self):
        return self.subclass_with_self(self.manager_cls)

    @cached_property
    def WorkerController(self):
        return self.subclass_with_self(self.worker_cls)

    @cached_property
    def Daemon(self):
        return self.subclass_with_self(self.daemon_cls)

    def register(self, *args, **options):
        """Register new handler."""

        def inner_register_handler(**options):

            def _register_handler(cls):
                # Check that handler is a class.
                if not inspect.isclass(cls):
                    raise RegistrationError('Object "{0!r}" is not a class'
                                            .format(cls))
                # Get processor for handler.
                processor = options.pop('processor', None)
                if processor is None:
                    raise RegistrationError('Processor for handler "{0!r}"'
                                            ' not specified'.format(cls))
                # Detect name for handler.
                name = options.pop('name', processor.__module__)
                self.slots.register(name=name,
                                    handler_cls=cls,
                                    processor_cls=processor,
                                    **options)
                return cls

            return _register_handler

        if len(args) == 1 and callable(args[0]):
            return inner_register_handler(**options)(*args)
        return inner_register_handler(**options)

    @cached_property
    def request_stack(self):
        """Store current requests."""
        return instantiate(self.request_stack_cls)



from __future__ import absolute_import

from functools import partial

from thriftworker.utils.loop import in_loop

from thriftpool.app import current_app


class Proxy(object):
    """Proxy object that execute functions on other side."""

    def __init__(self, producer):
        self.__producer = producer

    def __getattr__(self, name):
        """Create inner function that should enqueue remote procedure call."""
        producer = self.__producer

        def execute_callback(waiter, obj):
            if isinstance(obj, Exception):
                waiter.throw(obj)
            else:
                waiter.switch(obj)

        def inner_function(*args, **kwargs):
            waiter = current_app.hub.Waiter()
            producer.apply(name, callback=partial(execute_callback, waiter),
                           args=args, kwargs=kwargs)
            return waiter.get()

        inner_function.__name__ = name
        setattr(self, name, inner_function)
        return inner_function


class Client(object):
    """Client to """

    Proxy = Proxy

    def __init__(self, producer):
        self.__proxy = self.Proxy(producer)

    @in_loop
    def spawn(self, run, **kwargs):
        """Spawn given function in separate greenlet. Keyword
        argument `proxy` will be passed to provided function.

        """
        assert callable(run), 'given object not callable'
        kwargs.setdefault('proxy', self.__proxy)
        return current_app.hub.spawn(run, **kwargs)



from __future__ import absolute_import

import os
import signal
import logging

from pyuv import Pipe
from pyuv.errno import strerror

from thriftworker.utils.mixin import LoopMixin
from thriftworker.utils.loop import in_loop
from thriftworker.utils.decorators import cached_property

from thriftpool.components.base import StartStopComponent
from thriftpool.utils.mixin import LogsMixin

logger = logging.getLogger(__name__)


class ServicesManager(LogsMixin):

    def __init__(self, slots, services):
        self.slots = slots
        self.services = services
        super(ServicesManager, self).__init__()

    def start(self):
        for name in self.services:
            self.slots[name].start()

    def stop(self):
        for name in self.services:
            self.slots[name].stop()

    def register(self, name, processor):
        self._debug("Register service '%s'.", name)
        self.services.register(name, processor)


class ServicesComponent(StartStopComponent):

    name = 'worker.services'

    def create(self, parent):
        services = parent.app.thriftworker.services
        manager = ServicesManager(parent.app.slots, services)
        for slot in parent.app.slots:
            manager.register(slot.name, slot.service.processor)
        return manager


class WorkerComponent(StartStopComponent):

    name = 'worker.worker'
    requires = ('loop', 'services')

    def create(self, parent):
        return parent.app.thriftworker.worker


class AcceptorsComponent(StartStopComponent):

    name = 'worker.acceptors'
    requires = ('loop', 'worker')

    def create(self, parent):
        acceptors = parent.acceptors = parent.app.thriftworker.acceptors
        return acceptors


class Watchdog(LoopMixin):
    """Ensure that given socket alive or stop controller."""

    def __init__(self, app, descriptor):
        self.app = app
        self.descriptor = descriptor
        super(Watchdog, self).__init__()

    def _on_event(self, handle, data, error):
        if error:
            logger.error('Error %r happened on descriptor %d',
                         strerror(error), self.descriptor)
            self._pipe.close()
            # notify us that we should stop
            os.kill(os.getpid(), signal.SIGTERM)

    @cached_property
    def _pipe(self):
        pipe = Pipe(self.loop)
        pipe.open(self.descriptor)
        return pipe

    @in_loop
    def start(self):
        self._pipe.start_read(self._on_event)
        self._pipe.write('x')

    @in_loop
    def stop(self):
        if not self._pipe.closed:
            self._pipe.close()


class WatchdogComponent(StartStopComponent):

    name = 'worker.watchdog'
    requires = ('loop', )

    def create(self, parent):
        return Watchdog(parent.app, parent.handshake_fd)



"""Implement service repository."""
from __future__ import absolute_import

from collections import namedtuple

from thriftworker.utils.decorators import cached_property
from thriftworker.utils.imports import symbol_by_name, qualname

from thriftpool.request.handler import WrappedHandlerMeta
from thriftpool.request.processor import ProcessorMixin

__all__ = ['Repository']


class Listener(namedtuple('Listener', 'host port backlog')):
    """Specify which port we should listen."""


class ThriftService(namedtuple('ThriftService', ('service_name',
                                                 'processor_cls',
                                                 'handler_cls'))):
    """Describe service information."""

    def __reduce__(self):
        service_name, processor_cls, handler_cls = self
        processor_cls = qualname(processor_cls)
        handler_cls = qualname(handler_cls)
        return (self.__class__, (service_name, processor_cls, handler_cls))

    @cached_property
    def Handler(self):
        """Recreate handler class."""
        return symbol_by_name(self.handler_cls)

    @cached_property
    def handler(self):
        """Create handler instance."""
        return self.Handler()

    @cached_property
    def WrappedHandler(self):
        """Create wrapped handler instance."""
        Handler = self.Handler
        attrs = dict(_handler_cls=Handler,
                     _service_name=self.service_name)
        name = 'Wrapped{0}'.format(Handler.__name__)
        return WrappedHandlerMeta(name, (object, ), attrs)

    @cached_property
    def wrapped_handler(self):
        """Create wrapped handler instance."""
        return self.WrappedHandler(self.handler)

    @cached_property
    def Processor(self):
        """Create safe processor."""
        cls = symbol_by_name(self.processor_cls)
        attrs = dict(__module__=cls.__module__)
        return type(cls.__name__, (ProcessorMixin, cls), attrs)

    @cached_property
    def processor(self):
        """Create processor instance."""
        return self.Processor(self.wrapped_handler)


class Slot(namedtuple('Slot', 'name listener service')):
    """Combine service and listener together."""

    def __hash__(self):
        return hash(self.__class__) ^ hash(self.name)

    def start(self):
        try:
            initialize = self.service.handler.initialize
        except AttributeError:
            pass
        else:
            initialize()

    def stop(self):
        try:
            finalize = self.service.handler.finalize
        except AttributeError:
            pass
        else:
            finalize()


class Repository(set):
    """Store existed slots."""

    app = None

    Listener = Listener
    Service = ThriftService

    def __init__(self, slots=None):
        self._names = {}
        super(Repository, self).__init__()
        # Recreate repository from given slots.
        for slot in (slots or []):
            self.add(slot)

    def __reduce_args__(self):
        return (list(self),)

    def add(self, slot):
        """Add new slot to collection."""
        try:
            self._names[slot.name]
        except KeyError:
            self._names[slot.name] = slot
            super(Repository, self).add(slot)

    def __getitem__(self, name):
        """Get slot by name."""
        return self._names[name]

    def __contains__(self, name):
        """Check that given service registered in repository."""
        return name in self._names

    def register(self, name, processor_cls, handler_cls, **opts):
        """Register new service in repository."""
        # Create listener.
        listener = self.Listener(host=opts.get('host', '0.0.0.0'),
                                 port=opts.get('port'),
                                 backlog=opts.get('backlog'))
        # Create service.
        service = self.Service(service_name=name,
                               processor_cls=processor_cls,
                               handler_cls=handler_cls)
        # Create slot itself.
        slot = Slot(name, listener, service)
        self.add(slot)



"""Container for thrift services."""

VERSION = (0, 1, 23)

__version__ = '.'.join(map(str, VERSION[0:3]))
__author__ = 'Lipin Dmitriy'
__contact__ = 'blackwithwhite666@gmail.com'
__homepage__ = 'https://github.com/blackwithwhite666/thriftpool'
__docformat__ = 'restructuredtext'

# -eof meta-

from thriftpool.app._state import current_app as thriftpool, as_default_cls
from thriftpool.base import BaseHandler

__all__ = ['thriftpool', 'BaseHandler', 'as_default_cls']



"""Configure logging subsystem."""
from __future__ import absolute_import

from logging.handlers import WatchedFileHandler
import logging
import sys
import os

from thriftpool.signals import setup_logging, after_logging_setup
from thriftpool.utils.logs import ColorFormatter, LoggingProxy
from thriftpool.utils.term import colored, isatty
from thriftpool.utils.debug import RequestLogger

__all__ = ['Logging']


class Logging(object):
    """Setup logging subsystem."""

    app = None

    def __init__(self):
        config = self.app.config
        logfile = self.logfile = config.LOG_FILE
        self.loglevel = config.LOGGING_LEVEL
        self.format = config.DEFAULT_WORKER_LOG_FMT \
                      if os.getenv('IS_WORKER', False) \
                      else config.DEFAULT_LOG_FMT
        colorized = config.LOG_FORCE_COLORIZED = self.colorized = \
            config.LOG_FORCE_COLORIZED or self.colorize(logfile)
        self.colored = colored(enabled=colorized)
        self.request_logger = None
        self.tornado_logger = None

    def redirect_stdouts_to_logger(self, logger, loglevel=None, stdout=True,
                                   stderr=True):
        """Redirect :class:`sys.stdout` and :class:`sys.stderr` to a
        logging instance.

        :param logger: The :class:`logging.Logger` instance to redirect to.
        :param loglevel: The loglevel redirected messages will be logged as.

        """
        proxy = LoggingProxy(logger, loglevel)
        if stdout:
            sys.stdout = proxy
        if stderr:
            sys.stderr = proxy
        return proxy

    def get_handler(self, logfile=None):
        """Create log handler with either a filename, an open stream
        or :const:`None` (stderr).

        """
        logfile = sys.stderr if logfile is None else logfile
        if hasattr(logfile, 'write'):
            return logging.StreamHandler(logfile)
        return WatchedFileHandler(logfile)

    def colorize(self, logfile=None):
        """Can be output colored?"""
        return isatty(sys.stderr) if logfile is None else False

    def setup_handlers(self, logger, logfile, format,
                       formatter=ColorFormatter):
        """Register needed handlers for given logger. If ``logfile`` equal to
        :const:`None` use :attr:`sys.stderr` as ``logfile``.

        """
        colorize = self.colorize(logfile)
        datefmt = '%H:%M:%S' if logfile is None else '%Y-%m-%d %H:%M:%S'
        handler = self.get_handler(logfile)
        handler.setFormatter(formatter(format, use_color=colorize,
                                       datefmt=datefmt))
        logger.addHandler(handler)
        return logger

    def setup_request_logging(self):
        """Setup system to log requests."""
        logger = logging.getLogger('thriftpool.requests')
        self.request_logger = RequestLogger(logger, self.colored)
        self.request_logger.setup()

    def setup_tornado_logger(self):
        self.tornado_logger = logging.getLogger('thriftpool.tornado')

    def setup(self):
        root = logging.getLogger()
        root.setLevel(self.loglevel)
        receivers = setup_logging.send(sender=self, root=root,
                                       logfile=self.logfile,
                                       loglevel=self.loglevel)
        if not receivers:
            self.setup_handlers(root, self.logfile, self.format)
            after_logging_setup.send(sender=self, root=root,
                                     logfile=self.logfile,
                                     loglevel=self.loglevel)
        if self.app.config.LOG_REQUESTS:
            self.setup_request_logging()
        if self.app.config.LOG_TORNADO_REQUESTS:
            self.setup_tornado_logger()

    def log_tornado_request(self, handler):
        logger = self.tornado_logger
        if logger is None:
            return
        if handler.get_status() < 400:
            log_method = logger.info
        elif handler.get_status() < 500:
            log_method = logger.warning
        else:
            log_method = logger.error
        request_time = 1000.0 * handler.request.request_time()
        log_method("%d %s %.2fms", handler.get_status(),
                   handler._request_summary(), request_time)



"""Contains initializer for application."""
from __future__ import absolute_import

import importlib
from itertools import chain

try:
    from pkg_resources import iter_entry_points
except ImportError:
    iter_entry_points = lambda *args, **kwargs: []

from thriftworker.utils.imports import import_from_cwd

from thriftpool import signals

__all__ = ['Loader']


class Loader(object):
    """Provide configuration and some callback for main application."""

    app = None

    #: Specify modules that should be preloaded.
    builtin_modules = ['thriftpool.remote.handler']

    def get_config(self):
        """Return application configuration."""
        return {}

    def entrypoint_modules(self):
        """List all modules that are registered through setuptools
        entry points.

        """
        modules = []
        for entrypoint in iter_entry_points(group='thriftpool.modules'):
            # Grab the function that is the actual plugin.
            module_provider = entrypoint.load()
            modules.extend(module_provider(self.app))
        return modules

    def list_modules(self):
        """List all known module names."""
        for module_name in chain.from_iterable((self.builtin_modules,
                                                self.entrypoint_modules(),
                                                self.app.config.MODULES)):
            yield module_name

    def import_module(self, module):
        return import_from_cwd(module)

    def preload_modules(self):
        for module_name in self.list_modules():
            self.import_module(module_name)

    def on_before_init(self, controller):
        """Called before controller initialization."""
        pass

    def on_start(self):
        """Called before controller start."""
        signals.app_start.send(sender=self, app=self.app)

    def after_start(self):
        """Called after controller start."""
        signals.after_app_start.send(sender=self, app=self.app)

    def on_shutdown(self):
        """Called after controller shutdown."""
        signals.app_shutdown.send(sender=self, app=self.app)

    def after_unpickling(self):
        """Called after application was unpickled."""
        pass



from __future__ import absolute_import

from .base import StartStopComponent


class BaseLoopComponent(StartStopComponent):

    abstract = True

    def create(self, parent):
        return parent.app.thriftworker.hub


class WorkerLoopComponent(BaseLoopComponent):

    name = 'worker.loop'


class ManagerLoopComponent(BaseLoopComponent):

    name = 'manager.loop'



from __future__ import absolute_import

import time
from threading import Thread

from thriftpool.tests.utils import TestCase
from thriftpool.utils import local


class LocalTestCase(TestCase):

    def test_basic_local(self):
        l = local.Local()
        l.foo = 0
        values = []

        def value_setter(idx):
            time.sleep(0.01 * idx)
            l.foo = idx
            time.sleep(0.02)
            values.append(l.foo)

        threads = [Thread(target=value_setter, args=(x,))
                   for x in [1, 2, 3]]
        for thread in threads:
            thread.start()
        time.sleep(0.2)
        self.assertEqual(sorted(values), [1, 2, 3])

        def delfoo():
            del l.foo

        delfoo()
        with self.assertRaises(AttributeError):
            l.foo()
        with self.assertRaises(AttributeError):
            delfoo()

        local.release_local(l)

    def test_local_release(self):
        loc = local.Local()
        loc.foo = 42
        local.release_local(loc)
        self.assertFalse(hasattr(loc, 'foo'))

        ls = local.LocalStack()
        ls.push(42)
        local.release_local(ls)
        self.assertIsNone(ls.top)

    def test_local_proxy(self):
        foo = []
        ls = local.LocalProxy(lambda: foo)
        ls.append(42)
        ls.append(23)
        ls[1:] = [1, 2, 3]
        self.assertEqual(foo, [42, 1, 2, 3])
        self.assertEqual(repr(foo), repr(ls))
        self.assertEqual(foo[0], 42)
        foo += [1]
        self.assertEqual(list(foo), [42, 1, 2, 3, 1])

    def test_local_stack(self):
        ident = local.get_ident()

        ls = local.LocalStack()
        self.assertNotIn(ident, ls._local.__storage__)
        self.assertIsNone(ls.top)
        ls.push(42)
        self.assertIn(ident, ls._local.__storage__)
        self.assertEqual(ls.top, 42)
        ls.push(23)
        self.assertEqual(ls.top, 23)
        ls.pop()
        self.assertEqual(ls.top, 42)
        ls.pop()
        self.assertIsNone(ls.top)
        self.assertIsNone(ls.pop())
        self.assertIsNone(ls.pop())

        proxy = ls()
        ls.push([1, 2])
        self.assertEqual(proxy, [1, 2])
        ls.push((1, 2))
        self.assertEqual(proxy, (1, 2))
        ls.pop()
        ls.pop()
        self.assertEqual(repr(proxy), '<LocalProxy unbound>')

        self.assertNotIn(ident, ls._local.__storage__)

    def test_local_proxies_with_callables(self):
        foo = 42
        ls = local.LocalProxy(lambda: foo)
        self.assertEqual(ls, 42)
        foo = [23]
        ls.append(42)
        self.assertEqual(ls, [23, 42])
        self.assertEqual(foo, [23, 42])



import os
import re
import sys

from setuptools import setup, find_packages

# Check requirements.

if sys.version_info < (2, 7):
    raise Exception('ThriftPool requires Python 2.7.')


# Description, version and other meta information.

re_meta = re.compile(r'__(\w+?)__\s*=\s*(.*)')
re_vers = re.compile(r'VERSION\s*=\s*\((.*?)\)')
re_doc = re.compile(r'^"""(.+?)"""')
rq = lambda s: s.strip("\"'")


def add_default(m):
    attr_name, attr_value = m.groups()
    return ((attr_name, rq(attr_value)),)


def add_version(m):
    v = list(map(rq, m.groups()[0].split(', ')))
    return (('VERSION', '.'.join(v[0:3]) + ''.join(v[3:])),)


def add_doc(m):
    return (('doc', m.groups()[0]),)

pats = {re_meta: add_default,
        re_vers: add_version,
        re_doc: add_doc}
here = os.path.abspath(os.path.dirname(__file__))
meta_fh = open(os.path.join(here, 'thriftpool/__init__.py'))
try:
    meta = {}
    for line in meta_fh:
        if line.strip() == '# -eof meta-':
            break
        for pattern, handler in pats.items():
            m = pattern.match(line.strip())
            if m:
                meta.update(handler(m))
finally:
    meta_fh.close()

with open(os.path.join(here, 'README.rst')) as f:
    README = f.read()

with open(os.path.join(here, 'CHANGES.rst')) as f:
    CHANGES = f.read()


# Describe existed entry points

entrypoints = {}
entrypoints['console_scripts'] = [
    'thriftpoold = thriftpool.bin.thriftpoold:main',
    'thriftpool = thriftpool.bin.thriftpoolctl:main',
]
entrypoints['thriftpool.modules'] = []


setup(name='thriftpool',
      version=meta['VERSION'],
      description=meta['doc'],
      author=meta['author'],
      author_email=meta['contact'],
      url=meta['homepage'],
      long_description=README + '\n\n' + CHANGES,
      keywords='thrift soa',
      license='BSD',
      packages=find_packages(),
      install_requires=['thrift', 'pyuv>=0.8.3', 'six',
                        'gaffer>=0.4.4', 'thriftworker>=0.1.17'],
      entry_points=entrypoints,
      zip_safe=False
)



from __future__ import absolute_import

__all__ = ['SystemTerminate', 'RegistrationError']


class SystemTerminate(SystemExit):
    """Throw to terminate program."""


class RegistrationError(Exception):
    """Error happened on handler registration."""


class WrappingError(Exception):
    """Can't wrap given class."""



from __future__ import absolute_import

import logging
import cPickle as pickle
from uuid import UUID, uuid4
from struct import Struct
from cStringIO import StringIO

from gaffer.events import EventEmitter

from thriftworker.utils.decorators import cached_property
from thriftworker.utils.loop import in_loop

logger = logging.getLogger(__name__)


class Proto(object):
    """Mixin that produce *encode* and *decode* methods. Each packet
    represented by simple structure:

        +----------+--------------------+-----------------+
        | Length   | Request id (UUID)  | Pickled object  |
        +==========+====================+=================+
        | 4 bytes  | 32 bytes           | undefined       |
        +----------+--------------------+-----------------+

    """

    @cached_property
    def _length_struct(self):
        """Pack and unpack length."""
        return Struct('I')

    def _prepend_length(self, data):
        """Prepend length to data."""
        return self._length_struct.pack(len(data)) + data

    def _encode(self, request_id, obj):
        """Convert given request to packet."""
        assert isinstance(request_id, UUID), 'wrong request id given'
        return self._prepend_length(request_id.hex + pickle.dumps(obj))

    def _split_length(self, data):
        """Return tuple of length and raw data."""
        assert len(data) >= 4, 'given buffer to small (length)'
        encoded_length, left_data = data[:4], data[4:]
        length = self._length_struct.unpack(encoded_length)[0]
        assert length > 0, 'wrong length provided'
        return (length, left_data)

    def _split_request_id(self, data):
        """Return tuple of request id and raw data."""
        assert len(data) >= 32, 'given buffer to small (request_id)'
        request_id, left_data = data[:32], data[32:]
        return (UUID(hex=request_id, version=4), left_data)

    def _decode_body(self, data):
        """Unpickle given data."""
        return pickle.loads(data)


class Receiver(Proto):
    """Receive request from channel."""

    WAIT_LENGTH = 1
    WAIT_PAYLOAD = 2

    def __init__(self, stream, emitter):
        self.stream = stream
        self.emitter = emitter
        self._buf = StringIO()
        self._state = self.WAIT_LENGTH
        self._received = self._length = 0
        super(Receiver, self).__init__()

    def _on_read(self, evtype, info):
        data = info['data']
        if self._state == self.WAIT_LENGTH:
            # Try to receive packet length here.
            self._length, data = self._split_length(data)
            self._state = self.WAIT_PAYLOAD
        if self._state == self.WAIT_PAYLOAD:
            # Write data to buffer.
            self._received += len(data)
            self._buf.write(data)
        if self._received >= self._length:
            data = self._buf.getvalue()
            data, left = data[:self._length], data[self._length:]
            self._on_received(data)
            self._buf = StringIO()
            self._received = self._length = 0
            self._state = self.WAIT_LENGTH
            if left:
                self._on_read(evtype, dict(data=left))

    def _on_received(self, data):
        request_id, data = self._split_request_id(data)
        obj = self._decode_body(data)
        self.emitter.publish("received", request_id=request_id, obj=obj)

    def start(self):
        self.stream.subscribe(self._on_read)

    def stop(self):
        self.stream.unsubscribe(self._on_read)


class Transmitter(Proto):
    """Write request to channel."""

    def __init__(self, stream, emitter):
        self.stream = stream
        self.emitter = emitter
        self._requests = {}
        super(Transmitter, self).__init__()

    def write(self, obj, callback=None, request_id=None):
        request_id = request_id or uuid4()
        if callback is not None:
            self._requests[request_id] = callback
        data = self._encode(request_id, obj)
        self.stream.write(data)

    def _on_received(self, evtype, request_id, obj):
        callback = self._requests.pop(request_id, None)
        if callback is not None:
            try:
                callback(obj)
            except Exception as exc:
                logger.exception(exc)

    def start(self):
        self.emitter.subscribe('received', self._on_received)

    def stop(self):
        self.emitter.unsubscribe('received', self._on_received)


class Transport(object):
    """Exchange messages on channel."""

    Receiver = Receiver
    Transmitter = Transmitter

    def __init__(self, loop, incoming, outgoing):
        self.loop = loop
        self._emitter = EventEmitter(loop)
        self._receiver = self.Receiver(incoming, self._emitter)
        self._transmitter = self.Transmitter(outgoing, self._emitter)

    def write(self, obj, callback=None, request_id=None):
        self._transmitter.write(obj, callback, request_id)

    def subscribe(self, listener):
        self._emitter.subscribe('received', listener)

    def unsubscribe(self, listener):
        self._emitter.unsubscribe('received', listener)

    def start(self):
        self._receiver.start()
        self._transmitter.start()

    def stop(self):
        self._transmitter.stop()
        self._receiver.stop()
        self._emitter.close()


class Consumer(Transport):
    """Pull commands from consumer, execute them and return result."""

    def __init__(self, loop, incoming, outgoing, handler):
        self.handler = handler
        super(Consumer, self).__init__(loop, incoming, outgoing)

    def _on_incoming(self, evtype, request_id, obj):
        method_name, args, kwargs = obj
        logger.debug('Execute method {0!r} with args {1!r} and kwargs {2!r}'
                     .format(method_name, args, kwargs))
        try:
            result = getattr(self.handler, method_name)(*args, **kwargs)
        except Exception as exc:
            logger.exception(exc)
            result = exc
        self.write(result, request_id=request_id)

    def start(self):
        super(Consumer, self).start()
        self.subscribe(self._on_incoming)

    def stop(self):
        self.unsubscribe(self._on_incoming)
        super(Consumer, self).stop()


class Producer(Transport):
    """Push commands to consumer."""

    def __init__(self, loop, incoming, outgoing):
        super(Producer, self).__init__(loop, incoming, outgoing)

    @in_loop
    def apply(self, method_name, callback=None, args=None, kwargs=None):
        """Enqueue new remote procedure call."""
        self.write((str(method_name), args or [], kwargs or {}), callback)



from __future__ import absolute_import

import os
import sys
import socket

from thriftpool.utils.logs import mlevel, LOG_LEVELS, LEVELS
from thriftpool.utils.platforms import set_process_title
from thriftpool.bin.base import BaseCommand, Option


class ManagerCommand(BaseCommand):
    """Start ThiftPool daemon."""

    options = (
        Option('--log-request', help='Log all incoming requests',
               action='store_true'),
        Option('-c', '--concurrency', help='Set concurrency level',
               action='store', type=int),
        Option('-w', '--workers', help='Set workers count',
               action='store', type=int),
        Option('-k', '--worker-type', help='Set type of worker',
               action='store', type=str, choices=['sync', 'gevent']),
        Option('-l', '--log-level',
               help='Logging level', choices=LEVELS,
               action='store', default='INFO'),
        Option('-f', '--log-file', help='Specify log file',
               action='store'),
        Option('-p', '--pid-file', help='Specify path to PID file',
               action='store'),
        Option('-d', '--daemonize', help='Daemonize after start',
               action='store_true'),
        Option('-m', '--modules', help='Modules to load',
               action='store', type=str, nargs='*'),
        Option('--foreground', help='Don not detach from console',
               action='store_true'),
        Option('--endpoint', help='Which address tornado should listen?',
               action='store', type=str, nargs='*'),
    )

    def change_process_title(self, app):
        """Set process title."""
        set_process_title('[{0}@{1}]{2}'
                          .format(app.config.PROCESS_NAME,
                                  socket.gethostname(),
                                  ' '.join([''] + sys.argv[1:])))

    def run(self, *args, **options):
        app = self.app
        try:
            app.config.LOGGING_LEVEL = mlevel(options['log_level'])
        except KeyError:
            self.die('Unknown level {0!r}. Please use one of {1}.'
                     .format(options['log_level'],
                             '|'.join(LOG_LEVELS.keys())))

        normalize_path = lambda path: \
            os.path.abspath(os.path.expanduser(path)) \
            if path is not None else None

        pid_file = normalize_path(options.get('pid_file', None))
        log_file = normalize_path(options.get('log_file', None))

        app.config.LOG_REQUESTS = options.get('log_request', False)
        app.config.LOG_FILE = log_file

        if options['workers']:
            app.config.WORKERS = options['workers']
        if options['concurrency']:
            app.config.CONCURRENCY = options['concurrency']
        if options['worker_type']:
            app.config.WORKER_TYPE = options['worker_type']
        if options['modules']:
            modules = list(app.config.MODULES)
            modules.extend(options['modules'])
            app.config.MODULES = modules
        if options['endpoint']:
            modules = list(app.config.TORNADO_ENDPOINTS)
            modules.extend(options['endpoint'])
            app.config.TORNADO_ENDPOINTS = modules

        self.change_process_title(app)
        controller = app.ManagerController()
        daemon = app.Daemon(controller=controller,
                            pidfile=pid_file,
                            daemonize=options.get('daemonize', False),
                            foreground=options.get('foreground', False),
                            args=args)
        daemon.start()


def main():
    ManagerCommand().execute()


if __name__ == '__main__':
    main()



"""Some useful class for request logging."""
from __future__ import absolute_import

from functools import wraps
import time
import sys
import inspect
import itertools
from pprint import pformat

from thriftpool.signals import handler_method_guarded

if sys.platform == "win32":
    # On Windows, the best timer is time.clock()
    default_timer = time.clock
else:
    # On most other platforms the best timer is time.time()
    default_timer = time.time


NEW_REQUEST_MESSAGE = \
    """{prefix} do {method_name} where
    arguments = {arguments}
    keywords  = {keywords}"""

SERVED_REQUEST_MESSAGE = \
    """{prefix} return {result} ({took})"""


def qualname(obj):
    if not hasattr(obj, '__name__') and hasattr(obj, '__class__'):
        return qualname(obj.__class__)

    return '%s.%s' % (obj.__module__, obj.__name__)


class RequestLogger(object):

    def __init__(self, logger, colored):
        self.logger = logger
        self.colored = colored
        self.counter = itertools.cycle(xrange(2 ** 16))

    def setup(self):
        handler_method_guarded.connect(self.decorate)

    def decorate(self, signal, sender, fn):
        method_name = "{0}.{1}".format(qualname(sender), fn.__name__)
        method_args = inspect.getargspec(fn).args
        if inspect.ismethod(fn) and method_args:
            method_args.pop(0)
        blue = self.colored.blue
        black = self.colored.black
        magenta = self.colored.magenta

        def decorator(func):
            @wraps(func)
            def inner(*args, **kwargs):
                request = self.counter.next()
                # Log incoming request.
                arguments = []
                keywords = {}
                if method_args:
                    keywords.update(dict(zip(method_args[:len(args)], args)))
                else:
                    arguments = args
                keywords.update(kwargs)
                self.logger.info(NEW_REQUEST_MESSAGE.format(
                    prefix=magenta('In [{0}]:'.format(request)),
                    method_name=blue(method_name),
                    arguments=blue(pformat(arguments)),
                    keywords=blue(pformat(keywords)),
                ))
                # Measure time.
                start = default_timer()
                result = func(*args, **kwargs)
                duration = default_timer() - start
                # Log response.
                self.logger.info(SERVED_REQUEST_MESSAGE.format(
                    prefix=black('Out [{0}]:'.format(request)),
                    took='{:.3f} ms'.format(duration * 1000)
                         if duration < 0.001
                         else '{:.3f} s'.format(duration),
                    result=blue(pformat(result)),
                ))
                return result
            return inner

        return decorator
